import datetime
import uuid

import pytest

import globus_sdk
from globus_sdk.testing import get_last_request, load_response

ZERO_ID = uuid.UUID(int=0)


def get_last_params():
    return get_last_request().params


@pytest.mark.parametrize(
    "paramname, paramvalue",
    [
        ("filter_task_id", ZERO_ID),
        ("filter_task_id", "foo"),
        ("filter_owner_id", ZERO_ID),
        ("filter_owner_id", "foo"),
        ("filter_endpoint", ZERO_ID),
        ("filter_endpoint", "foo"),
        ("filter_is_paused", True),
        ("filter_is_paused", False),
        ("filter_min_faults", 0),
        ("filter_min_faults", 10),
        ("filter_local_user", "foouser"),
        ("filter_status", "ACTIVE"),
        ("filter_completion_time", "2020-08-25T00:00:00,2021-08-25T16:05:28"),
    ],
)
def test_strsafe_params(client, paramname, paramvalue):
    load_response(client.endpoint_manager_task_list)
    paramstr = str(paramvalue)
    client.endpoint_manager_task_list(**{paramname: paramvalue})
    params = get_last_params()
    assert paramname in params
    assert params[paramname] == paramstr


def test_filter_status_list(client):
    load_response(client.endpoint_manager_task_list)
    client.endpoint_manager_task_list(filter_status=["ACTIVE", "INACTIVE"])
    params = get_last_params()
    assert "filter_status" in params
    assert params["filter_status"] == "ACTIVE,INACTIVE"


def test_filter_task_id_list(client):
    load_response(client.endpoint_manager_task_list)
    # mixed list of str and UUID
    client.endpoint_manager_task_list(filter_task_id=["foo", ZERO_ID, "bar"])
    params = get_last_params()
    assert "filter_task_id" in params
    assert params["filter_task_id"] == f"foo,{str(ZERO_ID)},bar"


def test_filter_completion_time_datetime_tuple(client):
    load_response(client.endpoint_manager_task_list)

    dt1 = datetime.datetime.fromisoformat("2020-08-25T00:00:00")
    dt2 = datetime.datetime.fromisoformat("2021-08-25T16:05:28")

    client.endpoint_manager_task_list(filter_completion_time=(dt1, dt2))
    params = get_last_params()
    assert "filter_completion_time" in params
    assert params["filter_completion_time"] == "2020-08-25T00:00:00,2021-08-25T16:05:28"

    # mixed tuples work, important for passing `""`
    client.endpoint_manager_task_list(filter_completion_time=(dt1, ""))
    params = get_last_params()
    assert "filter_completion_time" in params
    assert params["filter_completion_time"] == "2020-08-25T00:00:00,"
    client.endpoint_manager_task_list(filter_completion_time=("", dt1))
    params = get_last_params()
    assert "filter_completion_time" in params
    assert params["filter_completion_time"] == ",2020-08-25T00:00:00"


@pytest.mark.parametrize("ep_use", ("source", "destination"))
def test_filter_by_endpoint_use(client, ep_use):
    meta = load_response(client.endpoint_manager_task_list).metadata
    if ep_use == "source":
        ep_id = meta["source"]
    else:
        ep_id = meta["destination"]

    client.endpoint_manager_task_list(filter_endpoint=ep_id, filter_endpoint_use=ep_use)
    params = get_last_params()

    assert "filter_endpoint" in params
    assert params["filter_endpoint"] == str(ep_id)
    assert "filter_endpoint_use" in params
    assert params["filter_endpoint_use"] == ep_use


@pytest.mark.parametrize("ep_use", ("source", "destination"))
def test_usage_error_on_filter_endpoint_use_without_endpoint(client, ep_use):
    with pytest.raises(
        globus_sdk.GlobusSDKUsageError,
        match=(
            "`filter_endpoint_use` is only valid when `filter_endpoint` is "
            r"also supplied\."
        ),
    ):
        client.endpoint_manager_task_list(filter_endpoint_use=ep_use)



from __future__ import annotations

import uuid

import pytest

from globus_sdk.testing import load_response


@pytest.mark.parametrize("uuid_type", (str, uuid.UUID))
def test_create_credential(
    service_client,
    uuid_type: type[str] | type[uuid.UUID],
):
    meta = load_response(service_client.create_client_credential).metadata

    res = service_client.create_client_credential(
        meta["client_id"] if uuid_type is str else uuid.UUID(meta["client_id"]),
        meta["name"],
    )

    assert res["credential"]["id"] == meta["credential_id"]


def test_create_credential_set_name(
    service_client,
):
    meta = load_response(service_client.create_client_credential, case="name").metadata

    res = service_client.create_client_credential(meta["client_id"], meta["name"])

    assert res["credential"]["name"] == meta["name"]



from __future__ import annotations

import logging
import typing as t

from globus_sdk._missing import MISSING, MissingType
from globus_sdk._payload import GlobusPayload

log = logging.getLogger(__name__)


class RunActivityNotificationPolicy(GlobusPayload):
    """
    A notification policy for a run, determining when emails will be sent.

    :param status: A list of statuses which will trigger notifications. When
        the run's status changes, it is evaluated against this list. If the new
        status matches the policy, an email is sent.
    """

    def __init__(
        self,
        status: (
            list[t.Literal["INACTIVE", "SUCCEEDED", "FAILED"]] | MissingType
        ) = MISSING,
    ) -> None:
        super().__init__()
        self["status"] = status



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

NEW_CREDENTIAL_NAME = str(uuid.uuid4()).replace("-", "")

CREDENTIAL = {
    "name": "foo",
    "id": str(uuid.uuid1()),
    "created": "2023-10-21T22:46:15.845937+00:00",
    "client": str(uuid.uuid1()),
    "secret": "abc123",
}


RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="auth",
        method="POST",
        path=f"/v2/api/clients/{CREDENTIAL['client']}/credentials",
        json={"credential": CREDENTIAL},
        metadata={
            "credential_id": CREDENTIAL["id"],
            "client_id": CREDENTIAL["client"],
            "name": CREDENTIAL["name"],
        },
    ),
    name=RegisteredResponse(
        service="auth",
        method="POST",
        path=f"/v2/api/clients/{CREDENTIAL['client']}/credentials",
        json={
            "credential": {
                **CREDENTIAL,
                "name": NEW_CREDENTIAL_NAME,
            }
        },
        metadata={
            "name": NEW_CREDENTIAL_NAME,
            "client_id": CREDENTIAL["client"],
        },
    ),
)



from globus_sdk.response import IterableResponse


class GetScopesResponse(IterableResponse):
    """
    Response class specific to the Get Scopes API

    Provides iteration on the "scopes" array in the response.
    """

    default_iter_key = "scopes"



from __future__ import annotations

import typing as t
import uuid

TWO_HOP_TRANSFER_FLOW_ID = "24bc4997-b483-4c25-a19c-64b0afc00743"
TWO_HOP_TRANSFER_FLOW_OWNER_ID = "b44bddda-d274-11e5-978a-9f15789a8150"

FLOW_ID = str(uuid.uuid1())
RUN_ID = str(uuid.uuid1())
USER1 = f"urn:globus:auth:identity:{uuid.uuid1()}"
USER2 = f"urn:globus:auth:identity:{uuid.uuid1()}"
GROUP = f"urn:globus:groups:id:{uuid.uuid1()}"
RUN_MANAGER_PRINCIPAL = f"urn:globus:auth:identity:{str(uuid.uuid1())}"
RUN_MONITOR_PRINCIPAL = f"urn:globus:auth:identity:{str(uuid.uuid1())}"

TWO_HOP_TRANSFER_FLOW_USER_SCOPE = (
    "https://auth.globus.org/scopes/"
    + TWO_HOP_TRANSFER_FLOW_ID
    + "/flow_"
    + TWO_HOP_TRANSFER_FLOW_ID.replace("-", "_")
    + "_user"
)

TWO_HOP_TRANSFER_FLOW_DEFINITION = {
    "States": {
        "Transfer1": {
            "Next": "Transfer2",
            "Type": "Action",
            "Comment": "Initial Transfer from Campus to DMZ",
            "ActionUrl": "https://actions.globus.org/transfer/transfer",
            "Parameters": {
                "transfer_items": [
                    {
                        "recursive": True,
                        "source_path.$": "$.source_path",
                        "destination_path.$": "$.staging_path",
                    }
                ],
                "source_endpoint_id.$": "$.source_endpoint_id",
                "destination_endpoint_id.$": "$.staging_endpoint_id",
            },
            "ResultPath": "$.Transfer1Result",
            "ActionScope": (
                "https://auth.globus.org/scopes/actions.globus.org/transfer/transfer"
            ),
        },
        "Transfer2": {
            "End": True,
            "Type": "Action",
            "Comment": "Transfer from DMZ to dataset repository",
            "ActionUrl": "https://actions.globus.org/transfer/transfer",
            "Parameters": {
                "transfer_items": [
                    {
                        "recursive": True,
                        "source_path.$": "$.staging_path",
                        "destination_path.$": "$.destination_path",
                    }
                ],
                "source_endpoint_id.$": "$.staging_endpoint_id",
                "destination_endpoint_id.$": "$.destination_endpoint_id",
            },
            "ResultPath": "$.Transfer2Result",
            "ActionScope": (
                "https://auth.globus.org/scopes/actions.globus.org/transfer/transfer"
            ),
        },
    },
    "Comment": "Two step transfer",
    "StartAt": "Transfer1",
}
TWO_HOP_TRANSFER_FLOW_DOC = {
    "id": TWO_HOP_TRANSFER_FLOW_ID,
    "definition": TWO_HOP_TRANSFER_FLOW_DEFINITION,
    "input_schema": {
        "type": "object",
        "required": [
            "source_endpoint_id",
            "source_path",
            "staging_endpoint_id",
            "staging_path",
            "destination_endpoint_id",
            "destination_path",
        ],
        "properties": {
            "source_path": {"type": "string"},
            "staging_path": {"type": "string"},
            "destination_path": {"type": "string"},
            "source_endpoint_id": {"type": "string"},
            "staging_endpoint_id": {"type": "string"},
            "destination_endpoint_id": {"type": "string"},
        },
        "additionalProperties": False,
    },
    "globus_auth_scope": TWO_HOP_TRANSFER_FLOW_USER_SCOPE,
    "synchronous": False,
    "log_supported": True,
    "types": ["Action"],
    "api_version": "1.0",
    "title": "Multi Step Transfer",
    "subtitle": "",
    "description": "",
    "keywords": ["two", "hop", "transfer"],
    "principal_urn": f"urn:globus:auth:identity:{TWO_HOP_TRANSFER_FLOW_ID}",
    "globus_auth_username": f"{TWO_HOP_TRANSFER_FLOW_ID}@clients.auth.globus.org",
    "created_at": "2020-09-01T17:59:20.711845+00:00",
    "updated_at": "2020-09-01T17:59:20.711845+00:00",
    "user_role": "flow_starter",
    "created_by": f"urn:globus:auth:identity:{TWO_HOP_TRANSFER_FLOW_OWNER_ID}",
    "visible_to": [],
    "runnable_by": [],
    "administered_by": [],
    "action_url": f"https://flows.globus.org/flows/{TWO_HOP_TRANSFER_FLOW_ID}",
    "flow_url": f"https://flows.globus.org/flows/{TWO_HOP_TRANSFER_FLOW_ID}",
    "flow_owner": f"urn:globus:auth:identity:{TWO_HOP_TRANSFER_FLOW_OWNER_ID}",
    "flow_viewers": [
        "public",
        "urn:globus:auth:identity:51abb9ce-6e05-4ab1-9a09-9c524313827c",
    ],
    "flow_starters": [
        "all_authenticated_users",
        "urn:globus:auth:identity:d0a15d1b-28f5-42de-9463-b8b6540421b6",
    ],
    "flow_administrators": [
        "urn:globus:auth:identity:05d29dab-bd26-4510-9290-468972e8ac01"
    ],
    "run_managers": [RUN_MANAGER_PRINCIPAL],
    "run_monitors": [RUN_MONITOR_PRINCIPAL],
}

TWO_HOP_TRANSFER_RUN_ID = "36ad9f9a-ad29-488f-beb4-c22ab729643a"
TWO_HOP_TRANSFER_RUN: dict[str, t.Any] = {
    "run_id": TWO_HOP_TRANSFER_RUN_ID,
    "flow_id": TWO_HOP_TRANSFER_FLOW_ID,
    "flow_title": TWO_HOP_TRANSFER_FLOW_DOC["title"],
    "flow_last_updated": "2020-09-01T17:59:20.711845+00:00",
    "start_time": "2020-09-12T15:00:20.711845+00:00",
    "status": "ACTIVE",
    "display_status": "ACTIVE",
    "details": {
        "code": "FlowStarting",
        "description": "The Flow is starting execution",
        "details": {
            "input": {
                "source_endpoint_id": "7e1b8ec7-a606-4c23-96c7-a2d930a3a55f",
                "source_path": "/path/to/the/source/dir",
                "staging_endpoint_id": "d5049dd6-ce9c-4f9e-853f-c25069f369f8",
                "staging_path": "/path/to/the/staging/dir",
                "destination_endpoint_id": "f3bd0daf-be5a-4df8-b53f-76b932113b7c",
                "destination_path": "/path/to/the/dest/dir",
            }
        },
    },
    "run_owner": TWO_HOP_TRANSFER_FLOW_DOC["flow_owner"],
    "run_managers": ["urn:globus:auth:identity:7d6064ef-5368-473a-b15b-e99c3561aa9b"],
    "run_monitors": [
        "urn:globus:auth:identity:58cf49f4-06ea-4b76-934c-d5c9f6c3ea9d",
        "urn:globus:auth:identity:57088a17-d5cb-4cfa-871a-c5cce48f2aec",
    ],
    "user_role": "run_owner",
    "label": "Transfer all of these files!",
    "tags": [
        "my-transfer-run",
        "jazz-fans",
    ],
    "search": {"task_id": "20ba91a8-eb90-470a-9477-2ad68808b276"},
}

FLOW_SCOPE_SUFFIX = f'flow_{FLOW_ID.replace("-", "_")}_user'
FLOW_SCOPE = f"https://auth.globus.org/scopes/{FLOW_ID}/{FLOW_SCOPE_SUFFIX}"
FLOW_DESCRIPTION = {
    "created_at": "2023-04-11T20:00:06.524930+00:00",
    "flow_owner": USER1,
    "created_by": USER1,
    "description": "This flow does some pretty cool stuff",
    "globus_auth_scope": FLOW_SCOPE,
    "id": FLOW_ID,
    "keywords": ["cool"],
    "subtitle": "My Cool Subtitle",
    "title": "My Cool Flow",
    "updated_at": "2023-04-11T20:00:06.524930+00:00",
}
RUN_DETAILS = {
    "code": "FlowSucceeded",
    "description": "The Flow run reached a successful completion state",
    "output": {
        "HelloResult": {
            "action_id": "6RxDm1JOQnG2",
            "completion_time": "2023-04-11T20:01:22.340594+00:00",
            "creator_id": USER1,
            "details": {"Hello": "World", "hello": "foo"},
            "display_status": "SUCCEEDED",
            "label": "My Cool Run",
            "manage_by": [USER2],
            "monitor_by": [GROUP],
            "release_after": None,
            "start_time": "2023-04-11T20:01:19.660251+00:00",
            "state_name": "RunHelloWorld",
            "status": "SUCCEEDED",
        },
        "input": {"echo_string": "foo", "sleep": 2},
    },
}
RUN = {
    "run_id": RUN_ID,
    "action_id": RUN_ID,
    "completion_time": "2023-04-11T20:01:22.917000+00:00",
    "created_by": USER1,
    "details": RUN_DETAILS,
    "display_status": "SUCCEEDED",
    "flow_id": FLOW_ID,
    "flow_last_updated": "2023-04-11T20:00:06.524930+00:00",
    "flow_title": "My Cool Flow",
    "label": "My Cool Run",
    "manage_by": [USER2],
    "monitor_by": [GROUP],
    "run_managers": [USER2],
    "run_monitors": [GROUP],
    "run_owner": USER1,
    "start_time": "2023-04-11T20:01:18.040416+00:00",
    "status": "SUCCEEDED",
    "tags": ["cool", "my"],
    "user_role": "run_owner",
}



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import ENDPOINT_ID, SUBSCRIPTION_ID

NO_ADMIN_ROLE_ENDPOINT_ID = str(uuid.UUID(int=10))
NON_SUBSCRIBED_ENDPOINT_ID = str(uuid.UUID(int=11))
NO_IDENTITIES_IN_SESSION_ENDPOINT_ID = str(uuid.UUID(int=12))


def _format_verify_route(epid: str) -> str:
    return f"/v0.10/endpoint/{epid}/subscription_admin_verified"


RESPONSES = ResponseSet(
    metadata={"endpoint_id": ENDPOINT_ID},
    default=RegisteredResponse(
        service="transfer",
        method="PUT",
        path=_format_verify_route(ENDPOINT_ID),
        status=200,
        json={
            "DATA_TYPE": "result",
            "code": "Updated",
            "message": "Endpoint updated successfully",
            "request_id": "SKWMqNWyv",
            "resource": _format_verify_route(ENDPOINT_ID),
        },
    ),
    no_admin_role=RegisteredResponse(
        metadata={"endpoint_id": NO_ADMIN_ROLE_ENDPOINT_ID},
        service="transfer",
        method="PUT",
        path=_format_verify_route(NO_ADMIN_ROLE_ENDPOINT_ID),
        status=403,
        json={
            "code": "PermissionDenied",
            "message": (
                "User does not have an admin role on the collection's "
                "subscription to set subscription_admin_verified"
            ),
            "request_id": "BHI2BHt8N",
            "resource": _format_verify_route(NO_ADMIN_ROLE_ENDPOINT_ID),
        },
    ),
    non_valid_verified_status=RegisteredResponse(
        service="transfer",
        method="PUT",
        path=_format_verify_route(ENDPOINT_ID),
        status=400,
        json={
            "code": "BadRequest",
            "message": (
                "Could not parse JSON: Expecting value: line 1 column 33 (char 32)"
            ),
            "request_id": "NPjnXpSD6",
            "resource": _format_verify_route(ENDPOINT_ID),
        },
    ),
    non_subscribed_endpoint=RegisteredResponse(
        metadata={"endpoint_id": NON_SUBSCRIBED_ENDPOINT_ID},
        service="transfer",
        method="PUT",
        path=_format_verify_route(NON_SUBSCRIBED_ENDPOINT_ID),
        status=400,
        json={
            "code": "BadRequest",
            "message": (
                "The collection must be associated with a subscription to "
                "set subscription_admin_verified"
            ),
            "request_id": "NPjnXpSD6",
            "resource": _format_verify_route(NON_SUBSCRIBED_ENDPOINT_ID),
        },
    ),
    no_identities_in_session=RegisteredResponse(
        metadata={
            "endpoint_id": NO_IDENTITIES_IN_SESSION_ENDPOINT_ID,
            "subscription_id": SUBSCRIPTION_ID,
        },
        service="transfer",
        method="PUT",
        path=_format_verify_route(NO_IDENTITIES_IN_SESSION_ENDPOINT_ID),
        status=400,
        json={
            "code": "BadRequest",
            "message": (
                "No manager or admin identities in session for high-assurance "
                f"subscription {SUBSCRIPTION_ID}"
            ),
            "request_id": "NPjnXpSD6",
            "resource": _format_verify_route(NO_IDENTITIES_IN_SESSION_ENDPOINT_ID),
        },
    ),
)



import pytest

import globus_sdk


@pytest.fixture
def auth_client():
    client = globus_sdk.ConfidentialAppAuthClient(
        "dummy_client_id", "dummy_client_secret"
    )
    with client.retry_config.tune(max_retries=0):
        yield client



import uuid

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

POLICY = {
    "high_assurance": False,
    "domain_constraints_include": ["greenlight.org"],
    "display_name": "GreenLight domain Only Policy",
    "description": "Only allow access from @greenlight.org",
    "id": str(uuid.uuid1()),
    "domain_constraints_exclude": None,
    "project_id": "da84e531-1afb-43cb-8c87-135ab580516a",
    "authentication_assurance_timeout": 35,
    "required_mfa": False,
}

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="auth",
        method="DELETE",
        path=f"/v2/api/policies/{POLICY['id']}",
        json={"policy": POLICY},
        metadata={
            "policy_id": POLICY["id"],
        },
    )
)



import base64
import hashlib
import os
from unittest import mock

import pytest

import globus_sdk
from globus_sdk.scopes import TransferScopes
from globus_sdk.services.auth.flow_managers.authorization_code import (
    GlobusAuthorizationCodeFlowManager,
)
from globus_sdk.services.auth.flow_managers.native_app import _make_native_app_challenge


@pytest.mark.parametrize(
    "verifier",
    [
        "x" * 20,  # too short
        "x" * 200,  # too long
        ("x" * 40) + "/" + ("y" * 40),  # includes invalid characters
    ],
)
def test_invalid_native_app_challenge(verifier):
    with pytest.raises(globus_sdk.GlobusSDKUsageError):
        _make_native_app_challenge(verifier)


def test_simple_input_native_app_challenge():
    verifier = "x" * 80
    challenge = (
        base64.urlsafe_b64encode(hashlib.sha256(verifier.encode("utf-8")).digest())
        .rstrip(b"=")
        .decode("utf-8")
    )
    res_verifier, res_challenge = _make_native_app_challenge(verifier)
    assert res_verifier == verifier
    assert res_challenge == challenge


def test_random_native_app_challenge(monkeypatch):
    b64vals = []

    def mock_urandom(n: int):
        return b"xyz"

    def mock_b64encode(b: bytes):
        b64vals.append(b)
        return b"abc123"

    monkeypatch.setattr(os, "urandom", mock_urandom)
    monkeypatch.setattr(base64, "urlsafe_b64encode", mock_b64encode)

    verifier, challenge = _make_native_app_challenge()
    assert verifier == "abc123"
    assert challenge == "abc123"

    assert len(b64vals) == 2
    assert b64vals == [b"xyz", hashlib.sha256(b"abc123").digest()]


def test_get_authorize_url_for_authorization_code():
    mock_client = mock.Mock()
    mock_client.client_id = "MOCK_CLIENT_ID"
    mock_client.base_url = "https://auth.globus.org/"
    flow_manager = GlobusAuthorizationCodeFlowManager(
        mock_client,
        redirect_uri="https://foo.example.org/authenticate",
        requested_scopes=TransferScopes.all,
    )

    silly_string = "ANANAS_IS_PINEAPPLE_BUT_BANANE_IS_BANANA"

    authorize_url = flow_manager.get_authorize_url()
    assert authorize_url.startswith("https://auth.globus.org")
    assert silly_string not in authorize_url

    silly_authorize_url = flow_manager.get_authorize_url(
        query_params={"silly_string": silly_string}
    )
    assert silly_string in silly_authorize_url



from __future__ import annotations

import sys
import typing as t
import uuid

# some error types use guards, so import from the specific module to avoid circularity
from globus_sdk.exc.base import ValidationError

if t.TYPE_CHECKING:
    from globus_sdk._internal.serializable import Serializable

if sys.version_info >= (3, 10):
    from typing import TypeGuard
else:
    from typing_extensions import TypeGuard

T = t.TypeVar("T")
S = t.TypeVar("S", bound="Serializable")


def is_list_of(data: t.Any, typ: type[T]) -> TypeGuard[list[T]]:
    return isinstance(data, list) and all(isinstance(item, typ) for item in data)


def is_optional(data: t.Any, typ: type[T]) -> TypeGuard[T | None]:
    return data is None or isinstance(data, typ)


def is_optional_list_of(data: t.Any, typ: type[T]) -> TypeGuard[list[T] | None]:
    return data is None or (
        isinstance(data, list) and all(isinstance(item, typ) for item in data)
    )


# this class is a namespace, separating validators (which error) from TypeGuards
class validators:
    @staticmethod
    def str_(name: str, value: t.Any) -> str:
        if isinstance(value, str):
            return value
        raise ValidationError(f"'{name}' must be a string")

    @staticmethod
    def int_(name: str, value: t.Any) -> int:
        if isinstance(value, int):
            return value
        raise ValidationError(f"'{name}' must be an int")

    @staticmethod
    def opt_str(name: str, value: t.Any) -> str | None:
        if is_optional(value, str):
            return value
        raise ValidationError(f"'{name}' must be a string or null")

    @staticmethod
    def opt_bool(name: str, value: t.Any) -> bool | None:
        if is_optional(value, bool):
            return value
        raise ValidationError(f"'{name}' must be a bool or null")

    @staticmethod
    def str_list(name: str, value: t.Any) -> list[str]:
        if is_list_of(value, str):
            return value
        raise ValidationError(f"'{name}' must be a list of strings")

    @staticmethod
    def opt_str_list(name: str, value: t.Any) -> list[str] | None:
        if is_optional_list_of(value, str):
            return value
        raise ValidationError(f"'{name}' must be a list of strings or null")

    @staticmethod
    def opt_str_list_or_commasep(name: str, value: t.Any) -> list[str] | None:
        if is_optional_list_of(value, str):
            return value
        if isinstance(value, str):
            return value.split(",")
        raise ValidationError(
            f"'{name}' must be a list of strings or a comma-delimited string or null"
        )

    @staticmethod
    def instance_or_dict(name: str, value: t.Any, cls: type[S]) -> S:
        if isinstance(value, cls):
            return value
        if isinstance(value, dict):
            return cls.from_dict(value)
        raise ValidationError(
            f"'{name}' must be a '{cls.__name__}' object or a dictionary"
        )

    @staticmethod
    def uuidlike(name: str, s: t.Any) -> uuid.UUID | str:
        """
        Raise an error if the input is not a UUID

        :param s: the UUID|str value
        :param name: the name for this value to use in error messages

        :raises TypeError: if the input is not a UUID|str
        :raises ValueError: if the input is a non-UUID str

        Example usage:

        .. code-block:: python

            def frob_it(collection_id: uuid.UUID | str) -> Frob:
                validators.uuidlike(collection_id, name="collection_id")
                return Frob(collection_id)

        """
        if isinstance(s, uuid.UUID):
            return s
        elif not isinstance(s, str):
            raise ValidationError(f"'{name}' must be a UUID or str (value='{s}')")

        try:
            uuid.UUID(s)
        except ValueError as e:
            raise ValidationError(f"'{name}' must be a valid UUID (value='{s}')") from e

        return s



import typing as t

from globus_sdk import MISSING, MissingType

# first, the type of `MISSING` must be `MissingType`
t.assert_type(MISSING, MissingType)

# second, a variable annotated as `int | MissingType` is assignable with `MISSING`
x: int | MissingType = MISSING

# and `MissingType` is not the same as None, Ellipsis, False, 0, or other weirdness
# these error!
y: MissingType
y = None  # type: ignore[assignment]
y = Ellipsis  # type: ignore[assignment]
y = ...  # type: ignore[assignment]
y = False  # type: ignore[assignment]
y = 0  # type: ignore[assignment]

# given that x is int|MissingType, `x is not MISSING` should narrow to `int`
if x is not MISSING:
    t.assert_type(x, int)
else:
    # don't do this:
    #   t.assert_type(x, MissingType)
    # although that looks right, `MissingType` != `Literal[MISSING]`, so it fails
    # (at least on some mypy versions)
    # instead, confirm that `not isinstance(x, MissingType)` narrows to a Never
    if not isinstance(x, MissingType):  # noqa: SDK003
        t.assert_never(x)



import uuid

import pytest

from globus_sdk.scopes import (
    AuthScopes,
    ComputeScopes,
    FlowsScopes,
    GCSCollectionScopes,
    GCSEndpointScopes,
    GroupsScopes,
    NexusScopes,
    Scope,
    SearchScopes,
    SpecificFlowScopes,
    TimersScopes,
    TransferScopes,
)


@pytest.mark.parametrize(
    "collection, expect_resource_server",
    (
        (AuthScopes, "auth.globus.org"),
        (ComputeScopes, "funcx_service"),
        (FlowsScopes, "flows.globus.org"),
        (GroupsScopes, "groups.api.globus.org"),
        (NexusScopes, "nexus.api.globus.org"),
        (SearchScopes, "search.api.globus.org"),
        (TimersScopes, "524230d7-ea86-4a52-8312-86065a9e0417"),
        (TransferScopes, "transfer.api.globus.org"),
    ),
)
def test_static_resource_server_attributes(collection, expect_resource_server):
    assert collection.resource_server == expect_resource_server


@pytest.mark.parametrize(
    "collection_cls", (GCSEndpointScopes, GCSCollectionScopes, SpecificFlowScopes)
)
def test_dynamic_resource_server_attributes(collection_cls):
    some_id = str(uuid.UUID(int=1))
    coll = collection_cls(some_id)
    assert coll.resource_server == some_id


def test_oidc_scope_formatting():
    assert str(AuthScopes.openid) == "openid"
    assert str(AuthScopes.email) == "email"
    assert str(AuthScopes.profile) == "profile"


def test_non_oidc_auth_scope_formatting():
    non_oidc_scopes = set(AuthScopes).difference(
        (AuthScopes.openid, AuthScopes.email, AuthScopes.profile)
    )

    assert len(non_oidc_scopes) > 0
    assert all(isinstance(x, Scope) for x in non_oidc_scopes)

    scope_strs = [str(s) for s in non_oidc_scopes]
    assert all(
        s.startswith("urn:globus:auth:scope:auth.globus.org:") for s in scope_strs
    )



import typing as t

import globus_sdk
from globus_sdk.scopes import Scope, ScopeParser
from globus_sdk.services.auth import (
    GlobusAuthorizationCodeFlowManager,
    GlobusNativeAppFlowManager,
)

# in tests below, passing `[None]` inline does not result in an arg-type error
# but instead in a list-item error because mypy effectively identifies that your
# argument type is "correct" (a list or iterable) and that it contains a
# bad item (the None value)
#
# therefore, define the none_list explicitly to control the error behavior better
none_list: list[None] = [None]

# setup clients for usage below
native_client = globus_sdk.NativeAppAuthClient("dummy_client_id")
cc_client = globus_sdk.ConfidentialAppAuthClient(
    "dummy_client_id", "dummy_client_secret"
)


# this function should type-check okay
def foo(x: str | Scope | t.Iterable[str | Scope]) -> str:
    return ScopeParser.serialize(x)


foo("somestring")
foo(["somestring", "otherstring"])
foo(Scope("bar"))
foo((Scope("bar"),))
foo({Scope("bar"), "baz"})
# bad usages
foo(1)  # type: ignore[arg-type]
foo((False,))  # type: ignore[arg-type]


# now, verify that we can pass scope collections to flow managers
GlobusAuthorizationCodeFlowManager(
    cc_client,
    "https://example.org/redirect-uri",
    requested_scopes="foo",
)
GlobusNativeAppFlowManager(
    native_client,
    requested_scopes="foo",
)
GlobusAuthorizationCodeFlowManager(
    cc_client,
    "https://example.org/redirect-uri",
    requested_scopes=("foo", "bar"),
)
GlobusNativeAppFlowManager(
    native_client,
    requested_scopes=("foo", "bar"),
)
GlobusAuthorizationCodeFlowManager(
    cc_client,
    "https://example.org/redirect-uri",
    requested_scopes=Scope("foo"),
)
GlobusNativeAppFlowManager(
    native_client,
    requested_scopes=Scope("foo"),
)
GlobusAuthorizationCodeFlowManager(
    cc_client,
    "https://example.org/redirect-uri",
    requested_scopes=[Scope("foo")],
)
GlobusNativeAppFlowManager(
    native_client,
    requested_scopes=[Scope("foo")],
)
GlobusAuthorizationCodeFlowManager(
    cc_client,
    "https://example.org/redirect-uri",
    requested_scopes=[Scope("foo"), "bar"],
)
GlobusNativeAppFlowManager(
    native_client,
    requested_scopes=[Scope("foo"), "bar"],
)
# bad usages
GlobusAuthorizationCodeFlowManager(
    cc_client,
    "https://example.org/redirect-uri",
    requested_scopes=1,  # type: ignore[arg-type]
)
GlobusNativeAppFlowManager(
    native_client,
    requested_scopes=1,  # type: ignore[arg-type]
)
GlobusAuthorizationCodeFlowManager(
    cc_client,
    "https://example.org/redirect-uri",
    requested_scopes=none_list,  # type: ignore[arg-type]
)
GlobusNativeAppFlowManager(
    native_client,
    requested_scopes=none_list,  # type: ignore[arg-type]
)


# furthermore, verify that we can pass these collection types to the client classes
# which wrap the flow managers
# note that oauth2_start_flow allows the scopes as a positional arg
native_client.oauth2_start_flow("foo")
cc_client.oauth2_start_flow("https://example.org/redirect-uri", "foo")
native_client.oauth2_start_flow(requested_scopes="foo")
cc_client.oauth2_start_flow("https://example.org/redirect-uri", requested_scopes="foo")
native_client.oauth2_start_flow(("foo", "bar"))
cc_client.oauth2_start_flow("https://example.org/redirect-uri", ("foo", "bar"))
native_client.oauth2_start_flow(requested_scopes=("foo", "bar"))
cc_client.oauth2_start_flow(
    "https://example.org/redirect-uri", requested_scopes=("foo", "bar")
)
native_client.oauth2_start_flow(Scope("foo"))
cc_client.oauth2_start_flow("https://example.org/redirect-uri", Scope("foo"))
native_client.oauth2_start_flow(requested_scopes=Scope("foo"))
cc_client.oauth2_start_flow(
    "https://example.org/redirect-uri", requested_scopes=Scope("foo")
)
native_client.oauth2_start_flow([Scope("foo"), "bar"])
cc_client.oauth2_start_flow("https://example.org/redirect-uri", [Scope("foo"), "bar"])
native_client.oauth2_start_flow(requested_scopes=[Scope("foo"), "bar"])
cc_client.oauth2_start_flow(
    "https://example.org/redirect-uri", requested_scopes=[Scope("foo"), "bar"]
)
# bad usages
native_client.oauth2_start_flow(1)  # type: ignore[arg-type]
cc_client.oauth2_start_flow(
    "https://example.org/redirect-uri",
    1,  # type: ignore[arg-type]
)
native_client.oauth2_start_flow(requested_scopes=1)  # type: ignore[arg-type]
cc_client.oauth2_start_flow(
    "https://example.org/redirect-uri",
    requested_scopes=1,  # type: ignore[arg-type]
)


# finally, requested_scopes for oauth2_client_credentials_tokens should follow these
# same constraints
cc_client.oauth2_client_credentials_tokens("foo")
cc_client.oauth2_client_credentials_tokens(requested_scopes="foo")
cc_client.oauth2_client_credentials_tokens(("foo", "bar"))
cc_client.oauth2_client_credentials_tokens(requested_scopes=("foo", "bar"))
cc_client.oauth2_client_credentials_tokens(Scope("foo"))
cc_client.oauth2_client_credentials_tokens(requested_scopes=Scope("foo"))
cc_client.oauth2_client_credentials_tokens([Scope("foo"), "bar"])
cc_client.oauth2_client_credentials_tokens(requested_scopes=[Scope("foo"), "bar"])
cc_client.oauth2_client_credentials_tokens(1)  # type: ignore[arg-type]
cc_client.oauth2_client_credentials_tokens(
    requested_scopes=none_list,  # type: ignore[arg-type]
)



from __future__ import annotations

import typing as t
import uuid

from globus_sdk import (
    AuthClient,
    AuthLoginClient,
    ConfidentialAppAuthClient,
    GlobusSDKUsageError,
    NativeAppAuthClient,
    Scope,
)
from globus_sdk.gare import GlobusAuthorizationParameters
from globus_sdk.login_flows import CommandLineLoginFlowManager, LoginFlowManager
from globus_sdk.token_storage import (
    HasRefreshTokensValidator,
    NotExpiredValidator,
    TokenStorage,
    UnchangingIdentityIDValidator,
    ValidatingTokenStorage,
)

from .app import GlobusApp
from .authorizer_factory import (
    AccessTokenAuthorizerFactory,
    RefreshTokenAuthorizerFactory,
)
from .config import DEFAULT_CONFIG, KNOWN_LOGIN_FLOW_MANAGERS, GlobusAppConfig
from .protocols import LoginFlowManagerProvider


class UserApp(GlobusApp):
    """
    A ``GlobusApp`` for managing authentication state of a user for use in service
    clients.

    Typically, a ``UserApp`` will use a native client, requiring a **client_id**
    created in a `Globus Project <https://app.globus.org/settings/developers>`_.
    More advanced use cases however, may additionally supply a **client_secret** or
    full **login_client** with confidential client credentials.

    ``UserApps`` are configured by supplying a :class:`GlobusAppConfig` object to the
    **config** parameter. Of note, login flow behavior involves printing and prompting
    the user for input using std::in and std::out. This behavior can be customized with
    the **login_flow_manager** config attribute.

    See :class:`GlobusApp` for method signatures.

    .. rubric:: Example Usage:

    .. code-block:: python

        app = UserApp("myapp", client_id=NATIVE_CLIENT_ID)
        transfer_client = TransferClient(app=app)
        res = transfer_client.endpoint_search("Tutorial Collection")

    :param app_name: A human-readable string to identify this app.
    :param login_client: A login client bound to a specific native client id or
        confidential client id/secret. Mutually exclusive with **client_id** and
        **client_secret**.
    :param client_id: A native or confidential client ID. Mutually exclusive with
        **login_client**.
    :param client_secret: A confidential client secret. Mutually exclusive with
        **login_client**.
    :param scope_requirements: A mapping of resource server to initial scope
        requirements.
    :param config: A data class containing configuration parameters for the app.
    """

    _login_client: NativeAppAuthClient | ConfidentialAppAuthClient
    _authorizer_factory: (  # type: ignore
        AccessTokenAuthorizerFactory | RefreshTokenAuthorizerFactory
    )

    def __init__(
        self,
        app_name: str = "Unnamed Globus App",
        *,
        login_client: AuthLoginClient | None = None,
        client_id: uuid.UUID | str | None = None,
        client_secret: str | None = None,
        scope_requirements: (
            t.Mapping[str, str | Scope | t.Iterable[str | Scope]] | None
        ) = None,
        config: GlobusAppConfig = DEFAULT_CONFIG,
    ) -> None:
        super().__init__(
            app_name,
            login_client=login_client,
            client_id=client_id,
            client_secret=client_secret,
            scope_requirements=scope_requirements,
            config=config,
        )

        self._login_flow_manager = self._resolve_login_flow_manager(
            app_name=self.app_name,
            login_client=self._login_client,
            config=config,
        )

    def _resolve_login_flow_manager(
        self, app_name: str, login_client: AuthLoginClient, config: GlobusAppConfig
    ) -> LoginFlowManager:
        login_flow_manager = config.login_flow_manager
        if isinstance(login_flow_manager, LoginFlowManager):
            return login_flow_manager

        elif isinstance(login_flow_manager, LoginFlowManagerProvider):
            provider = login_flow_manager
        elif login_flow_manager is None:
            provider = CommandLineLoginFlowManager
        elif login_flow_manager in KNOWN_LOGIN_FLOW_MANAGERS:
            provider = KNOWN_LOGIN_FLOW_MANAGERS[login_flow_manager]
        else:
            allowed_keys = ", ".join(repr(k) for k in KNOWN_LOGIN_FLOW_MANAGERS.keys())
            raise GlobusSDKUsageError(
                f"Unsupported login_flow_manager value: {login_flow_manager!r}. "
                f"Expected {allowed_keys}, a <LoginFlowManagerProvider>, or a "
                f"<LoginFlowManager>."
            )

        return provider.for_globus_app(
            app_name=app_name, config=config, login_client=login_client
        )

    def _initialize_login_client(
        self,
        app_name: str,
        config: GlobusAppConfig,
        client_id: uuid.UUID | str,
        client_secret: str | None,
    ) -> AuthLoginClient:
        if client_secret:
            return ConfidentialAppAuthClient(
                app_name=app_name,
                client_id=client_id,
                client_secret=client_secret,
                environment=config.environment,
            )
        else:
            return NativeAppAuthClient(
                app_name=app_name,
                client_id=client_id,
                environment=config.environment,
            )

    def _initialize_validating_token_storage(
        self,
        token_storage: TokenStorage,
        consent_client: AuthClient,
        scope_requirements: t.Mapping[str, t.Sequence[Scope]],
    ) -> ValidatingTokenStorage:
        validating_token_storage = super()._initialize_validating_token_storage(
            token_storage, consent_client, scope_requirements
        )
        validating_token_storage.validators.append(UnchangingIdentityIDValidator())
        return validating_token_storage

    def _initialize_authorizer_factory(self) -> None:
        if self.config.request_refresh_tokens:
            self._authorizer_factory = RefreshTokenAuthorizerFactory(
                token_storage=self.token_storage, auth_login_client=self._login_client
            )
            self.token_storage.validators.insert(0, HasRefreshTokensValidator())
        else:
            self._authorizer_factory = AccessTokenAuthorizerFactory(
                token_storage=self.token_storage
            )
            self.token_storage.validators.insert(0, NotExpiredValidator())

    def _run_login_flow(
        self, auth_params: GlobusAuthorizationParameters | None = None
    ) -> None:
        """
        Run an authorization flow to get new tokens which are stored and available
        for the next authorizer gotten by get_authorizer.

        As a UserApp this always involves an interactive login flow with the user
        driven by the app's LoginFlowManager.

        :param auth_params: A GlobusAuthorizationParameters to control how the user
            will authenticate. If not passed
        """
        auth_params = self._auth_params_with_required_scopes(auth_params)
        token_response = self._login_flow_manager.run_login_flow(auth_params)
        self._authorizer_factory.store_token_response_and_clear_cache(token_response)



from __future__ import annotations

import typing as t
import uuid
from unittest.mock import MagicMock, Mock

import pytest

import globus_sdk
from globus_sdk import (
    GlobusAppConfig,
    IDTokenDecoder,
    NativeAppAuthClient,
    OAuthTokenResponse,
    TransferClient,
    UserApp,
)
from globus_sdk.gare import GlobusAuthorizationParameters
from globus_sdk.login_flows import LoginFlowManager
from globus_sdk.scopes import Scope, TransferScopes
from globus_sdk.testing import RegisteredResponse
from globus_sdk.token_storage import MemoryTokenStorage, TokenStorageData


class GlobusAppConfigurator:

    def __init__(
        self,
        starting_tokens: dict[str, list[Scope]] | None = None,
        login_tokens: dict[str, list[Scope]] | None = None,
    ) -> None:
        self.identity_id = str(uuid.uuid4())

        self.token_storage = MemoryTokenStorage()
        if starting_tokens:
            self.token_storage.store_token_data_by_resource_server(
                {
                    resource_server: self._generate_token_data(resource_server, scopes)
                    for resource_server, scopes in starting_tokens.items()
                }
            )
        self._login_tokens: list[TokenStorageData] = [
            self._generate_token_data(resource_server, scopes)
            for resource_server, scopes in login_tokens.items() or {}.items()
        ]
        self.login_flow_manager = _FakeLoginFlowManager(self._login_tokens)

    def _generate_token_data(
        self, resource_server: str, scopes: list[Scope]
    ) -> TokenStorageData:
        return TokenStorageData(
            identity_id=self.identity_id,
            resource_server=resource_server,
            scope=" ".join(str(s) for s in scopes),
            access_token="generated-access-token",
            refresh_token="generated-refresh-token",
            expires_at_seconds=9999999999,
            token_type="Bearer",
        )

    def config(self, **kwargs: t.Any) -> GlobusAppConfig:
        decoder = MagicMock(spec=IDTokenDecoder)
        decoder.decode.return_value = {"sub": self.identity_id}

        return GlobusAppConfig(
            login_flow_manager=self.login_flow_manager,
            id_token_decoder=decoder,
            token_storage=self.token_storage,
            **kwargs,
        )


class _FakeLoginFlowManager(LoginFlowManager):

    def __init__(self, starting_tokens: list[TokenStorageData]) -> None:
        super().__init__(login_client=MagicMock(spec=NativeAppAuthClient))
        self._by_resource_server = {
            token.resource_server: token.to_dict() for token in starting_tokens
        }
        self.login_count = 0

    def run_login_flow(
        self, auth_parameters: GlobusAuthorizationParameters
    ) -> OAuthTokenResponse:
        self.login_count += 1
        response = Mock()
        response.id_token = "abcdefghjiklmnop"
        response.by_resource_server = self._by_resource_server
        return response


_GET_TASK_SUCCESS_RESPONSE = RegisteredResponse(
    service="transfer",
    path="/v0.10/task/foobar",
    method="GET",
    json={"task_id": "foobar"},
)

_GET_TASK_GARE_RESPONSE = RegisteredResponse(
    service="transfer",
    path="/v0.10/task/foobar",
    method="GET",
    status=403,
    json={
        "code": "AuthorizationRequired",
        "authorization_parameters": {"required_scopes": ["my-cool-new-scope"]},
    },
)


def test_app_can_redrive_gares():
    """
    When enabled:
    If an app-registered client encounters a GARE-compatible 403 http response, it
    should "redrive" it by:
    1. Running a login flow with the gare-included authorization
    2. Re-attempting the original request with the new token(s)
    """
    transfer_rs = TransferClient.resource_server
    configurator = GlobusAppConfigurator(
        # Start with a valid Transfer:all token
        starting_tokens={transfer_rs: [TransferScopes.all]},
        # On login, receive both Transfer:all and the gare-required scope
        login_tokens={transfer_rs: [TransferScopes.all, Scope("my-cool-new-scope")]},
    )
    config = configurator.config(auto_redrive_gares=True)

    app = UserApp(client_id="client_id", config=config)
    transfer = TransferClient(app=app)

    # Set up first a GARE response, then a successful response (on retry)
    _GET_TASK_GARE_RESPONSE.add()
    _GET_TASK_SUCCESS_RESPONSE.add()
    assert transfer.get_task("foobar").http_status == 200
    assert configurator.login_flow_manager.login_count == 1


def test_app_gare_redriving_is_disabled_by_default():
    transfer_rs = TransferClient.resource_server
    configurator = GlobusAppConfigurator(
        # Start with a valid Transfer:all token
        starting_tokens={transfer_rs: [TransferScopes.all]},
        # On login, receive both Transfer:all and the gare-required scope
        login_tokens={transfer_rs: [TransferScopes.all, Scope("my-cool-new-scope")]},
    )
    # Don't override the `auto_redrive_gares` default of False
    config = configurator.config()

    app = UserApp(client_id="client_id", config=config)
    transfer = TransferClient(app=app)

    # Set up first a GARE response, then a successful response (on retry)
    _GET_TASK_GARE_RESPONSE.add()
    _GET_TASK_SUCCESS_RESPONSE.add()
    with pytest.raises(globus_sdk.GlobusAPIError):
        transfer.get_task("foobar")
    assert configurator.login_flow_manager.login_count == 0


def test_app_gare_redrive_only_occurs_once_per_request():
    transfer_rs = TransferClient.resource_server
    configurator = GlobusAppConfigurator(
        # Start with a valid Transfer:all token
        starting_tokens={transfer_rs: [TransferScopes.all]},
        # On login, receive Transfer:all and a random scope (not required by the gare).
        login_tokens={transfer_rs: [TransferScopes.all, Scope("my-stupid-new-scope")]},
    )
    config = configurator.config(auto_redrive_gares=True)

    app = UserApp(client_id="client_id", config=config)
    transfer = TransferClient(app=app)

    # Set up exclusively a GARE response which will retry indefinitely.
    _GET_TASK_GARE_RESPONSE.add()
    with pytest.raises(globus_sdk.GlobusAPIError):
        transfer.get_task("foobar")
    assert configurator.login_flow_manager.login_count == 1



from __future__ import annotations

from globus_sdk.authorizers import GlobusAuthorizer

from .retry_config import RetryConfig


class RequestCallerInfo:
    """
    Data object that holds contextual information about the caller of a request.

    :param retry_config: The configuration of retry checks for the call
    :param resource_server: The resource server the request is being made to.
    :param authorizer: The authorizer object from the client making the request
    """

    def __init__(
        self,
        *,
        retry_config: RetryConfig,
        resource_server: str | None = None,
        authorizer: GlobusAuthorizer | None = None,
    ) -> None:
        self.resource_server = resource_server
        self.retry_config = retry_config
        self.authorizer = authorizer



#!/usr/bin/env python
"""
Meant to run in the context of a PR GitHub Actions workflow from the repo root
"""

import argparse
import glob
import json
import os
import subprocess
import typing as t
import urllib.error
import urllib.request

CHANGELOG_D = os.path.dirname(__file__)


def get_pr_number(commit_sha: str) -> t.Optional[str]:
    req = urllib.request.Request(
        "https://api.github.com/repos/globus/globus-sdk-python/"
        f"commits/{commit_sha}/pulls"
    )
    req.add_header("Accept", "application/vnd.github.v3+json")

    try:
        response = urllib.request.urlopen(req)
    except urllib.error.URLError:
        return None

    data = json.load(response)
    if not (isinstance(data, list) and len(data) > 0):
        return None
    prdata = data[0]
    if not (isinstance(prdata, dict) and "number" in prdata):
        return None
    return str(prdata["number"])


def update_file(fname: str, pr_num: t.Optional[str]) -> None:
    with open(fname) as f:
        content = f.read()

    if pr_num:
        content = content.replace(":pr:`NUMBER`", f":pr:`{pr_num}`")
    else:
        content = content.replace("(:pr:`NUMBER`)", "")
        content = "\n".join([line.rstrip() for line in content.splitlines()]) + "\n"

    with open(fname, "w") as f:
        f.write(content)


def get_commit(fname: str) -> str:
    return subprocess.run(
        ["git", "log", "-n", "1", "--pretty=format:%H", "--", fname],
        capture_output=True,
        check=True,
    ).stdout.decode("utf-8")


def main():
    parser = argparse.ArgumentParser(
        description="Update PR number in a changelog fragment"
    )
    parser.add_argument("FILENAME", help="file to check and update", nargs="?")
    args = parser.parse_args()

    if args.FILENAME:
        files = [args.FILENAME]
    else:
        files = glob.glob(os.path.join(CHANGELOG_D, "*.rst"))

    for filename in files:
        print(f"updating {filename}")
        commit = get_commit(filename)
        print(f"  commit={commit}")
        pr_num = get_pr_number(commit)
        print(f"  PR={pr_num}")
        update_file(filename, pr_num)


if __name__ == "__main__":
    main()



from ..collection import StaticScopeCollection, _urn_scope


class _TransferScopes(StaticScopeCollection):
    resource_server = "transfer.api.globus.org"

    all = _urn_scope(resource_server, "all")
    gcp_install = _urn_scope(resource_server, "gcp_install")


TransferScopes = _TransferScopes()



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import TIMER_ID, V1_TIMER

RESPONSES = ResponseSet(
    metadata={"job_id": TIMER_ID},
    default=RegisteredResponse(
        service="timer",
        path="/jobs/",
        method="POST",
        json=V1_TIMER,
        status=201,
    ),
    validation_error=RegisteredResponse(
        service="timer",
        path="/jobs/",
        method="POST",
        json={
            "detail": [
                {
                    "loc": ["body", "start"],
                    "msg": "field required",
                    "type": "value_error.missing",
                },
                {
                    "loc": ["body", "callback_url"],
                    "msg": "field required",
                    "type": "value_error.missing",
                },
            ]
        },
        metadata={
            "job_id": TIMER_ID,
            "expect_messages": [
                "field required: body.start",
                "field required: body.callback_url",
            ],
        },
        status=422,
    ),
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

REGISTERED_API_ID = "9dc867a4-1c6d-46d9-911f-71c2ac67e38c"
OWNER_URN = "urn:globus:auth:identity:a1234567-1234-1234-1234-123456789abc"
ADMIN_URN = "urn:globus:auth:identity:b2345678-2345-2345-2345-234567890abc"
VIEWER_URN = "urn:globus:auth:identity:c3456789-3456-3456-3456-345678901abc"

REGISTERED_API_DOC = {
    "id": REGISTERED_API_ID,
    "name": "test-registered-api",
    "description": "A test registered API for testing purposes",
    "roles": {
        "owners": [OWNER_URN],
        "administrators": [ADMIN_URN],
        "viewers": [VIEWER_URN],
    },
    "target": {
        "type": "openapi",
        "openapi_version": "3.1",
        "destination": {
            "method": "get",
            "url": "https://example.com/api/v1/resource/{resource_id}",
        },
        "specification": {
            "summary": "Get Resource",
            "description": "Retrieve a resource by ID",
            "deprecated": False,
            "parameters": [
                {
                    "name": "resource_id",
                    "in": "path",
                    "required": True,
                    "schema": {"type": "string"},
                    "allowEmptyValue": False,
                    "allowReserved": False,
                    "deprecated": False,
                }
            ],
            "responses": {
                "200": {
                    "description": "A resource response",
                    "content": {
                        "application/json": {
                            "schema": {"$ref": "#/components/schemas/ResourceResponse"}
                        }
                    },
                }
            },
            "security": [{"GlobusAuth": ["urn:globus:auth:scope:example.com:all"]}],
        },
        "transforms": None,
        "components": {
            "schemas": {
                "ResourceResponse": {
                    "type": "object",
                    "properties": {
                        "resource_id": {
                            "type": "string",
                            "format": "uuid",
                            "description": "The resource identifier",
                        },
                        "data": {
                            "type": "object",
                            "description": "The resource data",
                        },
                    },
                    "required": ["resource_id", "data"],
                    "additionalProperties": False,
                }
            }
        },
    },
    "data_templates": {
        "request": {
            "template": {
                "path": {"$T_ref": "$"},
            }
        },
        "response": {
            "2XX": {
                "template": {"$T_ref": "$.body"},
                "type": "success",
            },
            "default": {
                "template": {
                    "cause": {"$T_ref": "$.body"},
                    "exception": "ActionUnableToRun",
                },
                "type": "failure",
            },
        },
    },
    "state_input_schema": {
        "type": "object",
        "properties": {
            "resource_id": {"type": "string"},
        },
        "required": ["resource_id"],
    },
    "status": "ACTIVE",
    "subscription_id": "4fa609fc-14f6-4c62-acfd-680873d3b6fe",
    "created_timestamp": "2024-01-15T10:30:00+00:00",
    "edited_timestamp": None,
    "updated_timestamp": "2024-03-20T14:45:30+00:00",
    "scheduled_deletion_timestamp": None,
}

RESPONSES = ResponseSet(
    metadata={
        "registered_api_id": REGISTERED_API_ID,
        "name": REGISTERED_API_DOC["name"],
    },
    default=RegisteredResponse(
        service="flows",
        path=f"/registered_apis/{REGISTERED_API_ID}",
        json=REGISTERED_API_DOC,
    ),
)



import urllib.parse
import uuid

import pytest

import globus_sdk
from globus_sdk._missing import MISSING
from globus_sdk.scopes import TransferScopes
from globus_sdk.services.auth.flow_managers.native_app import _make_native_app_challenge
from globus_sdk.testing import load_response

CLIENT_ID = "d0f1d9b0-bd81-4108-be74-ea981664453a"


@pytest.fixture
def native_client():
    client = globus_sdk.NativeAppAuthClient(client_id=CLIENT_ID)
    with client.retry_config.tune(max_retries=0):
        yield client


@pytest.fixture
def confidential_client():
    client = globus_sdk.ConfidentialAppAuthClient(
        client_id=CLIENT_ID, client_secret="SECRET_SECRET_HES_GOT_A_SECRET"
    )
    with client.retry_config.tune(max_retries=0):
        yield client


# build a nearly-diagonal matrix over
#
#   domain: str | list[str]
#   identities: uuid | str | list[uuid] | list[str] | list[str | uuid]
#   policies: uuid | str | list[uuid] | list[str] | list[str | uuid]
#   mfa: True | False
#   message: str | None
#   prompt: Literal["prompt"] | None
#
# The order of these options is consequential.
# They are declared in the same order they tuple-unpack in the parametrized test.
# `None` values, though valid for each parameter, are omitted;
# they will be tested exactly once by a single all-`None` option test.
domain_options = ("example.edu", ["example.edu", "example.org"])
identity_options = (
    uuid.UUID(int=0),
    "foo-id",
    [uuid.UUID(int=0), uuid.UUID(int=1)],
    ["foo-id", "bar-id"],
    ["foo-id", uuid.UUID(int=2)],
)
policy_options = (
    uuid.UUID(int=3),
    "baz-id",
    [uuid.UUID(int=3), uuid.UUID(int=4)],
    ["baz-id", "quux-id"],
    ["baz-id", uuid.UUID(int=5)],
)
message_options = ("Test message",)
mfa_options = (True, False)
prompt_options = ("login",)
# Seed an all-`None` option test, then use a loop to fill in the rest.
# The number of parameters here must match the test parameters:
_ALL_SESSION_PARAM_COMBINATIONS = [(MISSING,) * 6]
for idx, options in enumerate(
    (
        domain_options,
        identity_options,
        policy_options,
        mfa_options,
        message_options,
        prompt_options,
    )
):
    for option in options:
        parameters = [MISSING] * 6
        parameters[idx] = option
        _ALL_SESSION_PARAM_COMBINATIONS.append(tuple(parameters))


@pytest.mark.parametrize("flow_type", ("native_app", "confidential_app"))
# parametrize over both what is and what *is not* passed as a parameter
@pytest.mark.parametrize(
    [
        "domain_option",
        "identity_option",
        "policy_option",
        "mfa_option",
        "message_option",
        "prompt_option",
    ],
    _ALL_SESSION_PARAM_COMBINATIONS,
)
def test_oauth2_get_authorize_url_supports_session_params(
    native_client,
    confidential_client,
    flow_type,
    domain_option,
    identity_option,
    policy_option,
    mfa_option,
    message_option,
    prompt_option,
):
    if flow_type == "native_app":
        client = native_client
    elif flow_type == "confidential_app":
        client = confidential_client
    else:
        raise NotImplementedError

    # get the url...
    client.oauth2_start_flow(redirect_uri="https://example.com", requested_scopes="foo")
    url_res = client.oauth2_get_authorize_url(
        session_required_single_domain=domain_option,
        session_required_identities=identity_option,
        session_required_policies=policy_option,
        session_required_mfa=mfa_option,
        session_message=message_option,
        prompt=prompt_option,
    )

    # parse the result...
    parsed_url = urllib.parse.urlparse(url_res)
    parsed_params = urllib.parse.parse_qs(parsed_url.query)

    # prepare some helper data...
    expected_params_keys = {
        "session_required_single_domain" if domain_option else None,
        "session_required_identities" if identity_option else None,
        "session_required_policies" if policy_option else None,
        "session_required_mfa" if mfa_option is not MISSING else None,
        "prompt" if prompt_option else None,
    }
    expected_params_keys.discard(None)
    unexpected_query_params = {
        "session_required_single_domain",
        "session_required_identities",
        "session_required_policies",
        "session_required_mfa",
        "prompt",
    } - expected_params_keys
    parsed_params_keys = set(parsed_params.keys())

    # ...and validate!
    assert expected_params_keys <= parsed_params_keys
    assert (unexpected_query_params - parsed_params_keys) == unexpected_query_params

    if domain_option is not MISSING:
        strized_option = (
            ",".join(str(x) for x in domain_option)
            if isinstance(domain_option, list)
            else str(domain_option)
        )
        assert parsed_params["session_required_single_domain"] == [strized_option]

    if identity_option is not MISSING:
        strized_option = (
            ",".join(str(x) for x in identity_option)
            if isinstance(identity_option, list)
            else str(identity_option)
        )
        assert parsed_params["session_required_identities"] == [strized_option]

    if policy_option is not MISSING:
        strized_option = (
            ",".join(str(x) for x in policy_option)
            if isinstance(policy_option, list)
            else str(policy_option)
        )
        assert parsed_params["session_required_policies"] == [strized_option]

    if mfa_option is not MISSING:
        strized_option = "True" if mfa_option else "False"
        assert parsed_params["session_required_mfa"] == [strized_option]

    if prompt_option is not MISSING:
        assert parsed_params["prompt"] == [prompt_option]


def test_oauth2_get_authorize_url_native_defaults(native_client):
    flow_manager = globus_sdk.services.auth.GlobusNativeAppFlowManager(
        native_client,
        TransferScopes.all,
    )
    native_client.current_oauth2_flow_manager = flow_manager

    # get url and validate results
    url_res = native_client.oauth2_get_authorize_url()
    parsed_url = urllib.parse.urlparse(url_res)
    assert f"https://{parsed_url.netloc}/" == native_client.base_url
    assert parsed_url.path == "/v2/oauth2/authorize"
    parsed_params = urllib.parse.parse_qs(parsed_url.query)
    assert parsed_params == {
        "client_id": [native_client.client_id],
        "redirect_uri": [native_client.base_url + "v2/web/auth-code"],
        "scope": [str(TransferScopes.all)],
        "state": ["_default"],
        "response_type": ["code"],
        "code_challenge": [flow_manager.challenge],
        "code_challenge_method": ["S256"],
        "access_type": ["online"],
    }


def test_oauth2_get_authorize_url_native_custom_params(native_client):
    # starting flow with custom parameters, should not warn because a scope is specified
    flow_manager = globus_sdk.services.auth.GlobusNativeAppFlowManager(
        native_client,
        requested_scopes="scopes",
        redirect_uri="uri",
        state="state",
        verifier=("a" * 43),
        refresh_tokens=True,
    )
    native_client.current_oauth2_flow_manager = flow_manager

    # get url_and validate results
    url_res = native_client.oauth2_get_authorize_url()
    verifier, remade_challenge = _make_native_app_challenge("a" * 43)
    parsed_url = urllib.parse.urlparse(url_res)
    assert f"https://{parsed_url.netloc}/" == native_client.base_url
    assert parsed_url.path == "/v2/oauth2/authorize"
    parsed_params = urllib.parse.parse_qs(parsed_url.query)
    assert parsed_params == {
        "client_id": [native_client.client_id],
        "redirect_uri": ["uri"],
        "scope": ["scopes"],
        "state": ["state"],
        "response_type": ["code"],
        "code_challenge": [urllib.parse.quote_plus(remade_challenge)],
        "code_challenge_method": ["S256"],
        "access_type": ["offline"],
    }


def test_oauth2_get_authorize_url_confidential_defaults(confidential_client):
    flow_manager = globus_sdk.services.auth.GlobusAuthorizationCodeFlowManager(
        confidential_client, "uri", TransferScopes.all
    )
    confidential_client.current_oauth2_flow_manager = flow_manager

    # get url_and validate results
    url_res = confidential_client.oauth2_get_authorize_url()
    parsed_url = urllib.parse.urlparse(url_res)
    assert f"https://{parsed_url.netloc}/" == confidential_client.base_url
    assert parsed_url.path == "/v2/oauth2/authorize"
    parsed_params = urllib.parse.parse_qs(parsed_url.query)
    assert parsed_params == {
        "client_id": [confidential_client.client_id],
        "redirect_uri": ["uri"],
        "scope": [str(TransferScopes.all)],
        "state": ["_default"],
        "response_type": ["code"],
        "access_type": ["online"],
    }


def test_oauth2_get_authorize_url_confidential_custom_params(confidential_client):
    # starting flow with specified parameters
    flow_manager = globus_sdk.services.auth.GlobusAuthorizationCodeFlowManager(
        confidential_client,
        requested_scopes="scopes",
        redirect_uri="uri",
        state="state",
        refresh_tokens=True,
    )
    confidential_client.current_oauth2_flow_manager = flow_manager

    # get url_and validate results
    url_res = confidential_client.oauth2_get_authorize_url()
    parsed_url = urllib.parse.urlparse(url_res)
    assert f"https://{parsed_url.netloc}/" == confidential_client.base_url
    assert parsed_url.path == "/v2/oauth2/authorize"
    parsed_params = urllib.parse.parse_qs(parsed_url.query)
    assert parsed_params == {
        "client_id": [confidential_client.client_id],
        "redirect_uri": ["uri"],
        "scope": ["scopes"],
        "state": ["state"],
        "response_type": ["code"],
        "access_type": ["offline"],
    }


def test_oauth2_exchange_code_for_tokens_native(native_client):
    """
    Starts a NativeAppFlowManager, Confirms invalid code raises 401
    Further testing cannot be done without user login credentials
    """
    load_response(native_client.oauth2_exchange_code_for_tokens, case="invalid_grant")

    flow_manager = globus_sdk.services.auth.GlobusNativeAppFlowManager(
        native_client, requested_scopes=TransferScopes.all
    )
    native_client.current_oauth2_flow_manager = flow_manager

    with pytest.raises(globus_sdk.AuthAPIError) as excinfo:
        native_client.oauth2_exchange_code_for_tokens("invalid_code")
    assert excinfo.value.http_status == 401
    assert excinfo.value.code is None


def test_oauth2_exchange_code_for_tokens_confidential(confidential_client):
    """
    Starts an AuthorizationCodeFlowManager, Confirms bad code raises 401
    Further testing cannot be done without user login credentials
    """
    load_response(
        confidential_client.oauth2_exchange_code_for_tokens, case="invalid_grant"
    )

    flow_manager = globus_sdk.services.auth.GlobusAuthorizationCodeFlowManager(
        confidential_client, "uri", requested_scopes=TransferScopes.all
    )
    confidential_client.current_oauth2_flow_manager = flow_manager

    with pytest.raises(globus_sdk.AuthAPIError) as excinfo:
        confidential_client.oauth2_exchange_code_for_tokens("invalid_code")
    assert excinfo.value.http_status == 401
    assert excinfo.value.code is None



import json

from globus_sdk import GCSRoleDocument
from globus_sdk.testing import get_last_request
from tests.common import register_api_route_fixture_file


def test_get_role_list(client):
    register_api_route_fixture_file("gcs", "/roles", "role_list.json")
    res = client.get_role_list()

    assert len(list(res)) == 2
    # sanity check some fields
    assert res["DATA_TYPE"] == "result#1.0.0"
    for item in res:
        assert item["DATA_TYPE"] == "role#1.0.0"
        assert "id" in item
        assert item["id"] in ("{role_id_1}", "{role_id_2}")
        assert "principal" in item
        assert "role" in item


def test_get_role_list_params(client):
    """
    confirms include, collection_id, and arbitrary query_params arguments
    to get_role_list are assembled correctly
    """
    register_api_route_fixture_file("gcs", "/roles", "role_list.json")

    # no args
    res = client.get_role_list()
    assert res["code"] == "success"
    params = get_last_request().params
    assert params == {}

    # collection_id
    res = client.get_role_list(collection_id="{collection_id_1}")
    assert res["code"] == "success"
    params = get_last_request().params
    assert params == {"collection_id": "{collection_id_1}"}

    # include
    res = client.get_role_list(include="all_roles")
    assert res["code"] == "success"
    params = get_last_request().params
    assert params == {"include": "all_roles"}

    # query_params
    res = client.get_role_list(query_params={"foo": "bar"})
    assert res["code"] == "success"
    params = get_last_request().params
    assert params == {"foo": "bar"}

    # everything together
    res = client.get_role_list(
        collection_id="{collection_id_1}",
        include="all_roles",
        query_params={"foo": "bar"},
    )
    assert res["code"] == "success"
    params = get_last_request().params
    assert params == {
        "collection_id": "{collection_id_1}",
        "include": "all_roles",
        "foo": "bar",
    }


def test_create_role(client):
    register_api_route_fixture_file(
        "gcs", "/roles", "role_document.json", method="POST"
    )

    data = GCSRoleDocument(
        collection="{collection_id_1}",
        principal="urn:globus:auth:identity:{user_id_1}",
        role="owner",
    )
    res = client.create_role(data)
    assert res["id"] == "{role_id_1}"

    json_body = json.loads(get_last_request().body)

    assert json_body["collection"] in (None, "{collection_id_1}")
    assert json_body["principal"] == "urn:globus:auth:identity:{user_id_1}"
    assert json_body["role"] in ("owner", "administrator")


def test_get_role(client):
    register_api_route_fixture_file("gcs", "/roles/ROLE_ID", "role_document.json")
    res = client.get_role("ROLE_ID")
    assert res["DATA_TYPE"] == "role#1.0.0"
    assert res["id"] == "{role_id_1}"
    assert res["collection"] is None
    assert res["principal"] == "urn:globus:auth:identity:{user_id_1}"
    assert res["role"] == "owner"


def test_delete_role(client):
    register_api_route_fixture_file(
        "gcs", "/roles/ROLE_ID", "empty_success.json", method="DELETE"
    )
    res = client.delete_role("ROLE_ID")
    assert res["DATA_TYPE"] == "result#1.0.0"
    assert "detail" in res.data
    assert res.data["detail"] == "success"



from .context import TokenValidationContext
from .errors import (
    ExpiredTokenError,
    IdentityMismatchError,
    IdentityValidationError,
    MissingIdentityError,
    MissingTokenError,
    TokenValidationError,
    UnmetScopeRequirementsError,
)
from .storage import ValidatingTokenStorage
from .validators import (
    HasRefreshTokensValidator,
    NotExpiredValidator,
    ScopeRequirementsValidator,
    TokenDataValidator,
    UnchangingIdentityIDValidator,
)

__all__ = (
    "TokenDataValidator",
    "TokenValidationContext",
    "ValidatingTokenStorage",
    "HasRefreshTokensValidator",
    "NotExpiredValidator",
    "ScopeRequirementsValidator",
    "UnchangingIdentityIDValidator",
    # errors
    "TokenValidationError",
    "IdentityValidationError",
    "IdentityMismatchError",
    "MissingIdentityError",
    "MissingTokenError",
    "ExpiredTokenError",
    "UnmetScopeRequirementsError",
)



import base64

import pytest

from globus_sdk.authorizers import BasicAuthorizer

USERNAME = "testUser"
PASSWORD = "PASSWORD"


@pytest.fixture
def authorizer():
    return BasicAuthorizer(USERNAME, PASSWORD)


def test_get_authorization_header(authorizer):
    """
    Gets authorization header, confirms expected value
    """
    header_val = authorizer.get_authorization_header()
    assert header_val[:6] == "Basic "
    assert header_val[6:] == "dGVzdFVzZXI6UEFTU1dPUkQ="
    decoded = base64.b64decode(header_val[6:].encode("utf-8")).decode("utf-8")
    assert decoded == f"{USERNAME}:{PASSWORD}"


def test_handle_missing_authorization(authorizer):
    """
    Confirms that BasicAuthorizer doesn't handle missing authorization
    """
    assert not authorizer.handle_missing_authorization()


@pytest.mark.parametrize(
    "username, password, encoded_value",
    [
        ("user", "テスト", "dXNlcjrjg4bjgrnjg4g="),
        ("дум", "pass", "0LTRg9C8OnBhc3M="),
        ("テスト", "дум", "44OG44K544OIOtC00YPQvA=="),
    ],
)
def test_unicode_handling(username, password, encoded_value):
    """
    With a unicode string for the password, set and verify the
    Authorization header.
    """
    authorizer = BasicAuthorizer(username, password)
    header_val = authorizer.get_authorization_header()

    assert header_val[:6] == "Basic "
    assert header_val[6:] == encoded_value
    decoded = base64.b64decode(header_val[6:].encode("utf-8")).decode("utf-8")
    assert decoded == f"{username}:{password}"



from globus_sdk.response import IterableResponse


class GetPoliciesResponse(IterableResponse):
    """
    Response class specific to the Get Policies API

    Provides iteration on the "policies" array in the response.
    """

    default_iter_key = "policies"



from copy import deepcopy

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import TWO_HOP_TRANSFER_FLOW_DOC, TWO_HOP_TRANSFER_FLOW_ID

_two_hop_transfer_update_request = {
    "subtitle": "Specifically, in two steps",
    "description": "Transfer from source to destination, stopping off at staging",
    "subscription_id": "00000000-3ba7-456e-9df7-fc40028f3331",
}


_updated_two_hop_transfer_flow_doc = deepcopy(TWO_HOP_TRANSFER_FLOW_DOC)
_updated_two_hop_transfer_flow_doc.update(_two_hop_transfer_update_request)


RESPONSES = ResponseSet(
    metadata={
        "flow_id": TWO_HOP_TRANSFER_FLOW_ID,
        "params": _two_hop_transfer_update_request,
    },
    default=RegisteredResponse(
        service="flows",
        path=f"/flows/{TWO_HOP_TRANSFER_FLOW_ID}",
        method="PUT",
        json=_updated_two_hop_transfer_flow_doc,
    ),
)



from globus_sdk.testing import load_response


def test_get_projects(service_client):
    meta = load_response(service_client.get_projects).metadata
    res = service_client.get_projects()

    assert [x["id"] for x in res] == meta["project_ids"]



from __future__ import annotations

import logging
import sys
import types
import typing as t
import urllib.parse

from globus_sdk import GlobusSDKUsageError, config, exc
from globus_sdk._internal.classprop import classproperty
from globus_sdk._internal.type_definitions import Closable
from globus_sdk._internal.utils import slash_join
from globus_sdk.authorizers import GlobusAuthorizer
from globus_sdk.paging import PaginatorTable
from globus_sdk.response import GlobusHTTPResponse
from globus_sdk.scopes import Scope, ScopeCollection
from globus_sdk.transport import RequestCallerInfo, RequestsTransport, RetryConfig
from globus_sdk.transport.default_retry_checks import DEFAULT_RETRY_CHECKS

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if t.TYPE_CHECKING:
    from globus_sdk.globus_app import GlobusApp

log = logging.getLogger(__name__)

_DataParamType: TypeAlias = t.Union[None, str, bytes, t.Dict[str, t.Any]]


class BaseClient:
    r"""
    Abstract base class for clients with error handling for Globus APIs.

    :param app: A ``GlobusApp`` which will be used for handling authorization and
        storing and validating tokens. Passing an ``app`` will automatically include
        a client's default scopes in the ``app``'s scope requirements unless specific
        ``app_scopes`` are given. If ``app_name`` is not given, the ``app``'s
        ``app_name`` will be used. Mutually exclusive with ``authorizer``.
    :param app_scopes: Optional list of ``Scope`` objects to be added to ``app``'s
        scope requirements instead of ``default_scope_requirements``. Requires ``app``.
    :param authorizer: A ``GlobusAuthorizer`` which will generate Authorization headers.
        Mutually exclusive with ``app``.
    :param app_name: Optional "nice name" for the application. Has no bearing on the
        semantics of client actions. It is just passed as part of the User-Agent
        string, and may be useful when debugging issues with the Globus Team.
        If both``app`` and ``app_name`` are given, this value takes priority.
    :param base_url: The URL for the service. Most client types initialize this value
        intelligently by default. Set it when inheriting from BaseClient or
        communicating through a proxy. This value takes precedence over the class
        attribute of the same name.
    :param transport: A :class:`RequestsTransport` object for sending and
        retrying requests. By default, one will be constructed by the client.
    :param retry_config: A :class:`RetryConfig` object with parameters to
        control request retry behavior. By default, one will be constructed by
        the client.
    """

    # service name is used to lookup a service URL from config
    service_name: str = "_base"

    # the URL for the service
    # NOTE: this is not the only way to define a base url. See the docstring of the
    # `BaseClient._resolve_base_url` method for more details.
    base_url: str = "_base"

    #: the class for errors raised by this client on HTTP 4xx and 5xx errors
    #: this can be set in subclasses, but must always be a subclass of GlobusError
    error_class: type[exc.GlobusAPIError] = exc.GlobusAPIError

    #: the scopes for this client may be present as a ``ScopeCollection``
    scopes: ScopeCollection | None = None

    def __init__(
        self,
        *,
        environment: str | None = None,
        base_url: str | None = None,
        app: GlobusApp | None = None,
        app_scopes: list[Scope] | None = None,
        authorizer: GlobusAuthorizer | None = None,
        app_name: str | None = None,
        transport: RequestsTransport | None = None,
        retry_config: RetryConfig | None = None,
    ) -> None:
        # check for input parameter conflicts
        if app_scopes and not app:
            raise exc.GlobusSDKUsageError(
                f"A {type(self).__name__} must have an 'app' to use 'app_scopes'."
            )
        if app and authorizer:
            raise exc.GlobusSDKUsageError(
                f"A {type(self).__name__} cannot use both an 'app' and an 'authorizer'."
            )

        self._resources_to_close: list[Closable] = []

        # Determine the client's environment
        # Either the provided kwarg or derived from the app used
        #
        # If neither is specified, fallback to the GLOBUS_SDK_ENVIRONMENT environment
        # variable.
        if environment:
            self.environment = environment
        elif app:
            self.environment = app.config.environment
        else:
            self.environment = config.get_environment_name()

        # resolve the base_url for the client (see docstring for resolution precedence)
        self.base_url = self._resolve_base_url(base_url, self.environment)

        self.retry_config: RetryConfig = retry_config or RetryConfig()
        self._register_standard_retry_checks(self.retry_config)

        # the client owns the Transport, and is responsible for closing on close,
        # if and only if the client creates the Transport
        if transport is not None:
            self.transport = transport
        else:
            self.transport = RequestsTransport()
            self._resources_to_close.append(self.transport)

        log.debug(f"initialized transport of type {type(self.transport)}")

        # setup paginated methods
        self.paginated = PaginatorTable(self)

        # set application name if available from app_name
        # if this is not set, `app.app_name` may be applied below
        self._app_name: str | None = None
        if app_name is not None:
            self.app_name = app_name

        # attach the app or authorizer provided
        # starting app attributes as `None` and calling the attachment method
        self.authorizer = authorizer
        self._app: GlobusApp | None = None
        self.app_scopes: list[Scope] | None = None
        if app:
            self.attach_globus_app(app, app_scopes=app_scopes)

    @property
    def default_scope_requirements(self) -> list[Scope]:
        """
        Scopes that will automatically be added to this client's app's
        scope_requirements during _finalize_app.

        For clients with static scope requirements this can just be a static
        value. Clients with dynamic requirements should use @property and must
        return sane results while the Base Client is being initialized.
        """
        raise NotImplementedError

    def _register_standard_retry_checks(self, retry_config: RetryConfig) -> None:
        """
        Setup the standard checks for this client.

        This is called during init and may be overridden by subclasses.
        """
        retry_config.checks.register_many_checks(DEFAULT_RETRY_CHECKS)

    @classmethod
    def _resolve_base_url(cls, init_base_url: str | None, environment: str) -> str:
        """
        Resolve the client's base url.

        Precedence (this evaluation will fall through if an option is not set):
          1. [Highest] Constructor `base_url` value.
          2. Class `base_url` attribute.
          3. Class `service_name` attribute (computed).

        :param init_base_url: The `base_url` value supplied to the constructor.
        :param environment: The environment to use for service URL resolution.
        :returns: The resolved base URL.
        :raises: GlobusSDKUsageError if base_url cannot be resolved.
        """
        if init_base_url is not None:
            log.debug(f"Creating client of type {cls}")
            return init_base_url
        elif cls.base_url != "_base":
            log.debug(f"Creating client of type {cls}")
            return cls.base_url
        elif cls.service_name != "_base":
            log.debug(f'Creating client of type {cls} for service "{cls.service_name}"')
            return config.get_service_url(cls.service_name, environment)

        raise GlobusSDKUsageError(
            f"Unable to resolve base_url in client {cls}. "
            f"Clients must define either one or both of 'base_url' and 'service_name'."
        )

    def attach_globus_app(
        self, app: GlobusApp, app_scopes: list[Scope] | None = None
    ) -> None:
        """
        Attach a ``GlobusApp`` to this client and, conversely, register this client with
        that app. The client's default scopes will be added to the app's scope
        requirements unless ``app_scopes`` is used to override this.

        If the ``app_name`` is not set on the client, it will be set to match that of
        the app.

        .. note::

            This method is only safe to call once per client object. It is implicitly
            called if the client is initialized with an app.

        :param app: The ``GlobusApp`` to attach to this client.
        :param app_scopes: Optional list of ``Scope`` objects to be added to ``app``'s
            scope requirements instead of ``default_scope_requirements``. These will be
            stored in the ``app_scopes`` attribute of the client.

        :raises GlobusSDKUsageError: If the attachment appears to conflict with the
            state of the client. e.g., an app or authorizer is already in place.
        """
        # If there are any incompatible or ambiguous data, usage error.
        # "In the face of ambiguity, refuse the temptation to guess."
        if self._app:
            raise exc.GlobusSDKUsageError(
                f"Cannot attach GlobusApp to {type(self).__name__} when one is "
                "already attached."
            )
        if self.app_scopes:
            # technically, we *could* allow for this, but it's not clear what
            # it would mean if a user wrote the following:
            #
            #   c = ClientClass()
            #   c.app_scopes = [foo]
            #   c.attach_globus_app(app, app_scopes=[bar])
            #
            # did the user expect a merge, overwrite, or other behavior?
            raise exc.GlobusSDKUsageError(
                f"Cannot attach GlobusApp to {type(self).__name__} when `app_scopes` "
                "is already set. "
                "The scopes for this client cannot be consistently resolved."
            )
        if self.authorizer:
            raise exc.GlobusSDKUsageError(
                f"Cannot attach GlobusApp to {type(self).__name__} when it "
                "has an authorizer assigned."
            )
        if self.resource_server is None:
            raise exc.GlobusSDKUsageError(
                "Unable to use an 'app' with a client with no "
                "'resource_server' defined."
            )
        # the client's environment must match the app's
        #
        # there are only two ways to get to a mismatch:
        #
        # 1. pass an explicit environment which doesn't match the app, e.g.,
        #   `MyClient(environment="a", app=app)` where `app.config.environment="b"`
        #
        # 2. initialize a client without an app and later attach an app which doesn't
        #    match, e.g., `MyClient(environment="a").attach_globus_app(app)`
        #
        # in these cases, the user has explicitly given us conflicting instructions
        if self.environment != app.config.environment:
            raise exc.GlobusSDKUsageError(
                f"[Environment Mismatch] {type(self).__name__}'s environment "
                f"({self.environment}) does not match the GlobusApp's configured "
                f"environment ({app.config.environment})."
            )

        # now, assign the app, app_name, and scopes
        self._app = app
        self.app_scopes = app_scopes or self.default_scope_requirements
        if self.app_name is None:
            self.app_name = app.app_name

        # register the scope requirements on the app side
        self._app.add_scope_requirements({self.resource_server: self.app_scopes})

        # finally, set up auto-gare redriving by attaching any necessary retry checks.
        self.retry_config.checks.register_many_checks(app.get_client_retry_checks())

    def add_app_scope(
        self, scope_collection: str | Scope | t.Iterable[str | Scope]
    ) -> BaseClient:
        """
        Add a given scope collection to this client's ``GlobusApp`` scope requirements
        for this client's ``resource_server``. This allows defining additional scope
        requirements beyond the client's ``default_scope_requirements``.

        Returns ``self`` for chaining.

        Raises ``GlobusSDKUsageError`` if this client was not initialized with a
            ``GlobusApp``.

        :param scope_collection: A scope or scopes of
            ``str | Scope | t.Iterable[str | Scope]`` to be added to
            the app's required scopes.

        .. tab-set::

        .. tab-item:: Example Usage

            .. code-block:: python

                app = UserApp("myapp", ...)
                flows_client = (
                    FlowsClient(app=app)
                    .add_app_scope(FlowsScopes.manage_flows)
                    .add_app_scope(FlowsScopes.run_manage)
                )

        """
        if not self._app:
            raise exc.GlobusSDKUsageError(
                "Cannot 'add_app_scope' on a client that does not have an 'app'."
            )
        if self.resource_server is None:
            raise ValueError(
                "Unable to use an 'app' with a client with no "
                "'resource_server' defined."
            )
        self._app.add_scope_requirements({self.resource_server: scope_collection})

        return self

    @property
    def app_name(self) -> str | None:
        return self._app_name

    @app_name.setter
    def app_name(self, value: str) -> None:
        self._app_name = self.transport.user_agent = value

    @classproperty
    def resource_server(  # pylint: disable=missing-param-doc
        self_or_cls: BaseClient | type[BaseClient],
    ) -> str | None:
        """
        The resource_server name for the API and scopes associated with this client.

        This information is pulled from the ``scopes`` attribute of the client class.
        If the client does not have associated scopes, this value will be ``None``.

        This must return sane results while the Base Client is being initialized.
        """
        if self_or_cls.scopes is None:
            return None
        return self_or_cls.scopes.resource_server

    def close(self) -> None:
        """
        Close all resources which are owned by this client.

        This only closes transports which are created implicitly via client init.
        Externally constructed transports will not be closed.
        """
        for resource in self._resources_to_close:
            log.debug(
                f"closing resource of type {type(resource).__name__} "
                f"for {type(self).__name__}"
            )
            resource.close()

    # clients can act as context managers, and such usage calls close()
    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> None:
        self.close()

    def get(  # pylint: disable=missing-param-doc
        self,
        path: str,
        *,
        query_params: dict[str, t.Any] | None = None,
        headers: dict[str, str] | None = None,
        automatic_authorization: bool = True,
    ) -> GlobusHTTPResponse:
        """
        Make a GET request to the specified path.

        See :py:meth:`~.BaseClient.request` for details on the various parameters.
        """
        log.debug(f"GET to {path} with query_params {query_params}")
        return self.request(
            "GET",
            path,
            query_params=query_params,
            headers=headers,
            automatic_authorization=automatic_authorization,
        )

    def post(  # pylint: disable=missing-param-doc
        self,
        path: str,
        *,
        query_params: dict[str, t.Any] | None = None,
        data: _DataParamType = None,
        headers: dict[str, str] | None = None,
        encoding: str | None = None,
        automatic_authorization: bool = True,
    ) -> GlobusHTTPResponse:
        """
        Make a POST request to the specified path.

        See :py:meth:`~.BaseClient.request` for details on the various parameters.
        """
        log.debug(f"POST to {path} with query_params {query_params}")
        return self.request(
            "POST",
            path,
            query_params=query_params,
            data=data,
            headers=headers,
            encoding=encoding,
            automatic_authorization=automatic_authorization,
        )

    def delete(  # pylint: disable=missing-param-doc
        self,
        path: str,
        *,
        query_params: dict[str, t.Any] | None = None,
        headers: dict[str, str] | None = None,
        automatic_authorization: bool = True,
    ) -> GlobusHTTPResponse:
        """
        Make a DELETE request to the specified path.

        See :py:meth:`~.BaseClient.request` for details on the various parameters.
        """
        log.debug(f"DELETE to {path} with query_params {query_params}")
        return self.request(
            "DELETE",
            path,
            query_params=query_params,
            headers=headers,
            automatic_authorization=automatic_authorization,
        )

    def put(  # pylint: disable=missing-param-doc
        self,
        path: str,
        *,
        query_params: dict[str, t.Any] | None = None,
        data: _DataParamType = None,
        headers: dict[str, str] | None = None,
        encoding: str | None = None,
        automatic_authorization: bool = True,
    ) -> GlobusHTTPResponse:
        """
        Make a PUT request to the specified path.

        See :py:meth:`~.BaseClient.request` for details on the various parameters.
        """
        log.debug(f"PUT to {path} with query_params {query_params}")
        return self.request(
            "PUT",
            path,
            query_params=query_params,
            data=data,
            headers=headers,
            encoding=encoding,
            automatic_authorization=automatic_authorization,
        )

    def patch(  # pylint: disable=missing-param-doc
        self,
        path: str,
        *,
        query_params: dict[str, t.Any] | None = None,
        data: _DataParamType = None,
        headers: dict[str, str] | None = None,
        encoding: str | None = None,
        automatic_authorization: bool = True,
    ) -> GlobusHTTPResponse:
        """
        Make a PATCH request to the specified path.

        See :py:meth:`~.BaseClient.request` for details on the various parameters.
        """
        log.debug(f"PATCH to {path} with query_params {query_params}")
        return self.request(
            "PATCH",
            path,
            query_params=query_params,
            data=data,
            headers=headers,
            encoding=encoding,
            automatic_authorization=automatic_authorization,
        )

    def request(
        self,
        method: str,
        path: str,
        *,
        query_params: dict[str, t.Any] | None = None,
        data: _DataParamType = None,
        headers: dict[str, str] | None = None,
        encoding: str | None = None,
        allow_redirects: bool = True,
        stream: bool = False,
        automatic_authorization: bool = True,
    ) -> GlobusHTTPResponse:
        """
        Send an HTTP request

        :param method: HTTP request method, as an all caps string
        :param path: Path for the request, with or without leading slash
        :param query_params: Parameters to be encoded as a query string
        :param headers: HTTP headers to add to the request. Authorization headers may
            be overwritten unless ``automatic_authorization`` is False.
        :param data: Data to send as the request body. May pass through encoding.
        :param encoding: A way to encode request data. "json", "form", and "text"
            are all valid values. Custom encodings can be used only if they are
            registered with the transport. By default, strings get "text" behavior and
            all other objects get "json".
        :param allow_redirects: Follow Location headers on redirect response
            automatically. Defaults to ``True``
        :param stream: Do not immediately download the response content. Defaults to
            ``False``
        :param automatic_authorization: Use this client's ``app`` or ``authorizer``
            to automatically generate an Authorization header.

        :raises GlobusAPIError: a `GlobusAPIError` will be raised if the response to the
            request is received and has a status code in the 4xx or 5xx categories
        """
        # prepare data...
        # copy headers if present
        rheaders = {**headers} if headers else {}

        # if a client is asked to make a request against a full URL, not just the path
        # component, then do not resolve the path, simply pass it through as the URL
        if path.startswith(("https://", "http://")):
            url = path
        else:
            url = slash_join(self.base_url, urllib.parse.quote(path))

        caller_info = RequestCallerInfo(retry_config=self.retry_config)

        # either use given authorizer or get one from app
        if automatic_authorization:
            caller_info.authorizer = self.authorizer
            if self._app and (resource_server := self.resource_server):
                caller_info.resource_server = resource_server
                caller_info.authorizer = self._app.get_authorizer(resource_server)

        # make the request
        log.debug("request will hit URL: %s", url)
        r = self.transport.request(
            method,
            url,
            caller_info=caller_info,
            data=data,
            query_params=query_params,
            headers=rheaders,
            encoding=encoding,
            allow_redirects=allow_redirects,
            stream=stream,
        )
        log.debug("request made to URL: %s", r.url)

        if 200 <= r.status_code < 400:
            log.debug(f"request completed with response code: {r.status_code}")
            return GlobusHTTPResponse(r, self)

        log.debug(f"request completed with (error) response code: {r.status_code}")
        raise self.error_class(r)



from __future__ import annotations

import abc
import sys
import typing as t

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

# TODO: Remove this dispatch after we drop Python 3.8 support.
#       In 3.9+ `dict.__class_getitem__` is available.
if t.TYPE_CHECKING:
    # pylint: disable=unsubscriptable-object
    _PayloadBaseDict = t.Dict[str, t.Any]
else:
    _PayloadBaseDict = dict


class GlobusPayload(_PayloadBaseDict):
    """
    A class for defining helper objects which wrap some kind of "payload" dict.
    Typical for helper objects which formulate a request payload.

    Payload types inheriting from this class can be passed directly to the client
    ``post()``, ``put()``, and ``patch()`` methods. These methods will
    recognize a ``PayloadBase`` and apply conversions for serialization with
    the requested encoder (e.g. as a JSON request body).
    """


class AbstractGlobusPayload(GlobusPayload, abc.ABC):
    """
    An abstract class which is a GlobusPayload.

    This is a shim which is needed because we have a metaclass conflict between
    dict:type and ABC:ABCMeta.

    Setting the metaclass helps type checkers understand that such classes are
    abstract.
    """

    # explicitly define `__new__` in order to check for abstract methods which
    # were not redefined
    def __new__(cls, *args: t.Any, **kwargs: t.Any) -> Self:
        obj = super().__new__(cls, *args, **kwargs)
        abstractmethods: frozenset[str] = (
            obj.__abstractmethods__  # type: ignore[attr-defined]
        )
        if abstractmethods:
            s = "" if len(abstractmethods) == 1 else "s"
            methodnames = ", ".join(f"'{f}'" for f in abstractmethods)
            # this error very closely imitates the errors produced by ABCMeta
            raise TypeError(
                f"Can't instantiate abstract class {cls.__name__} without "
                f"an implementation for abstract method{s} {methodnames}"
            )
        return obj



from __future__ import annotations

import pytest

from globus_sdk.testing import load_response


@pytest.mark.parametrize(
    "case_name",
    (
        "client_id_str",
        "client_id_uuid",
        "name",
        "description",
        "scope_suffix",
        "no_required_domains",
        "required_domains",
        "no_dependent_scopes",
        "dependent_scopes",
        "advertised",
        "not_advertised",
        "allows_refresh_token",
        "disallows_refresh_token",
    ),
)
def test_create_scope(
    service_client,
    case_name: str,
):
    meta = load_response(service_client.create_scope, case=case_name).metadata

    res = service_client.create_scope(**meta["args"])
    for k, v in meta["response"].items():
        assert res["scope"][k] == v



from globus_sdk.testing import get_last_request, load_response


def test_oauth2_token_introspect(auth_client):
    meta = load_response(auth_client.oauth2_token_introspect).metadata
    response = auth_client.oauth2_token_introspect("some_very_cool_token")
    assert response["username"] == meta["username"]
    assert response["sub"] == meta["id"]

    last_req = get_last_request()
    assert len(last_req.params) == 0


def test_oauth2_token_introspect_allows_added_query_params(auth_client):
    meta = load_response(auth_client.oauth2_token_introspect).metadata
    response = auth_client.oauth2_token_introspect(
        "some_very_cool_token", query_params={"foo": "bar"}
    )
    assert response["username"] == meta["username"]
    assert response["sub"] == meta["id"]

    last_req = get_last_request()
    assert last_req.params["foo"] == "bar"



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import SUBMISSION_ID

RESPONSES = ResponseSet(
    metadata={"submission_id": SUBMISSION_ID},
    default=RegisteredResponse(
        service="transfer",
        path="/v0.10/submission_id",
        json={"value": SUBMISSION_ID},
    ),
)



from unittest.mock import Mock

import pytest

from globus_sdk.login_flows import LocalServerLoginError
from globus_sdk.login_flows.local_server_login_flow_manager.local_server import (  # noqa: E501
    DEFAULT_HTML_TEMPLATE,
    RedirectHandler,
    RedirectHTTPServer,
)


def test_default_html_template_contains_expected_text():
    # basic integrity test
    assert "<h1>Globus Login Result</h1>" in DEFAULT_HTML_TEMPLATE.substitute(
        post_login_message="", login_result="Login successful"
    )


@pytest.mark.parametrize(
    "url,expected_result",
    [
        (b"localhost?code=abc123", "abc123"),
        (b"localhost?error=bad_login", LocalServerLoginError("bad_login")),
        (b"localhost", LocalServerLoginError(None)),
    ],
)
def test_server(url, expected_result):
    """
    Setup a RedirectHTTPServer and pass it mocked HTTP GET requests to have
    its RedirectHandler handle
    """
    server = RedirectHTTPServer(
        server_address=("", 0),
        handler_class=RedirectHandler,
        html_template=DEFAULT_HTML_TEMPLATE,
    )

    # setup Mocks to look like a connection to a file for reading the HTTP data
    mock_file = Mock()
    mock_file.readline.side_effect = [
        b"GET " + url + b" HTTP/1.1",
        b"Host: localhost",
        b"",
    ]
    mock_conn = Mock()
    mock_conn.makefile.return_value = mock_file

    # handle the request, then cleanup
    server.finish_request(mock_conn, ("", 0))
    server.server_close()

    # confirm expected results
    result = server.wait_for_code()
    if isinstance(result, str):
        assert result == expected_result
    elif isinstance(result, LocalServerLoginError):
        assert result.args == expected_result.args
    else:
        raise AssertionError("unexpected result type")



from __future__ import annotations

import logging
import typing as t
import uuid

from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey

from globus_sdk import client, exc
from globus_sdk._internal.remarshal import commajoin
from globus_sdk._missing import MISSING, MissingType
from globus_sdk.authorizers import GlobusAuthorizer, NullAuthorizer
from globus_sdk.response import GlobusHTTPResponse
from globus_sdk.scopes import AuthScopes, Scope
from globus_sdk.transport import RequestsTransport, RetryConfig

from .._common import get_jwk_data, pem_decode_jwk_data
from ..errors import AuthAPIError
from ..flow_managers import GlobusOAuthFlowManager
from ..response import (
    OAuthAuthorizationCodeResponse,
    OAuthRefreshTokenResponse,
    OAuthTokenResponse,
)

log = logging.getLogger(__name__)

RT = t.TypeVar("RT", bound=GlobusHTTPResponse)


class AuthLoginClient(client.BaseClient):
    """
    This client class provides the common base for clients providing login
    functionality via
    `Globus Auth <https://docs.globus.org/api/auth/>`_

    :param client_id: The ID of the application provided by registration with
        Globus Auth.

    All other initialization parameters are passed through to ``BaseClient``.

    .. automethodlist:: globus_sdk.AuthLoginClient
    """

    service_name = "auth"
    error_class = AuthAPIError
    scopes = AuthScopes

    def __init__(
        self,
        client_id: uuid.UUID | str | None = None,
        environment: str | None = None,
        base_url: str | None = None,
        authorizer: GlobusAuthorizer | None = None,
        app_name: str | None = None,
        transport: RequestsTransport | None = None,
        retry_config: RetryConfig | None = None,
    ) -> None:
        super().__init__(
            environment=environment,
            base_url=base_url,
            authorizer=authorizer,
            app_name=app_name,
            transport=transport,
            retry_config=retry_config,
        )
        self.client_id: str | None = str(client_id) if client_id is not None else None
        # an AuthClient may contain a GlobusOAuth2FlowManager in order to
        # encapsulate the functionality of various different types of flow
        # managers
        self.current_oauth2_flow_manager: GlobusOAuthFlowManager | None = None

        log.debug(
            "Finished initializing AuthLoginClient. "
            f"client_id='{client_id}', type(authorizer)={type(authorizer)}"
        )

    @property
    def default_scope_requirements(self) -> list[Scope]:
        raise exc.GlobusSDKUsageError(
            "AuthLoginClients do not have default_scope_requirements, "
            "use AuthClient instead."
        )

    # FYI: this get_openid_configuration method is duplicated in AuthClient
    # if this code is modified, please update that copy as well
    #
    # we would like to restructure code using this method to be calling the matching
    # AuthClient method
    # for example, a future SDK version may make an AuthLoginClient contain
    # an AuthClient which it uses
    def get_openid_configuration(self) -> GlobusHTTPResponse:
        """
        Fetch the OpenID Connect configuration data from the well-known URI for Globus
        Auth.
        """
        log.debug("Fetching OIDC Config")
        return self.get("/.well-known/openid-configuration")

    @t.overload
    def get_jwk(
        self,
        openid_configuration: GlobusHTTPResponse | dict[str, t.Any] | MissingType,
        *,
        as_pem: t.Literal[True],
    ) -> RSAPublicKey: ...

    @t.overload
    def get_jwk(
        self,
        openid_configuration: GlobusHTTPResponse | dict[str, t.Any] | MissingType,
        *,
        as_pem: t.Literal[False],
    ) -> dict[str, t.Any]: ...

    # FYI: this get_jwk method is duplicated in AuthClient
    # if this code is modified, please update that copy as well
    #
    # we would like to restructure code using this method to be calling the matching
    # AuthClient method
    # for example, a future SDK version may make an AuthLoginClient contain
    # an AuthClient which it uses
    def get_jwk(
        self,
        openid_configuration: (
            GlobusHTTPResponse | dict[str, t.Any] | MissingType
        ) = MISSING,
        *,
        as_pem: bool = False,
    ) -> RSAPublicKey | dict[str, t.Any]:
        """
        Fetch the Globus Auth JWK.

        Returns either a dict or an RSA Public Key object depending on ``as_pem``.

        :param openid_configuration: The OIDC config as a GlobusHTTPResponse or dict.
            When not provided, it will be fetched automatically.
        :param as_pem: Decode the JWK to an RSA PEM key, typically for JWT decoding
        """
        if openid_configuration is MISSING:
            log.debug("No OIDC Config provided, autofetching...")
            openid_configuration = self.get_openid_configuration()
        jwk_data = get_jwk_data(
            fget=self.get, openid_configuration=openid_configuration
        )
        return pem_decode_jwk_data(jwk_data=jwk_data) if as_pem else jwk_data

    def oauth2_get_authorize_url(
        self,
        *,
        session_required_identities: (
            uuid.UUID | str | t.Iterable[uuid.UUID | str] | MissingType
        ) = MISSING,
        session_required_single_domain: str | t.Iterable[str] | MissingType = MISSING,
        session_required_policies: (
            uuid.UUID | str | t.Iterable[uuid.UUID | str] | MissingType
        ) = MISSING,
        session_required_mfa: bool | MissingType = MISSING,
        session_message: str | MissingType = MISSING,
        prompt: t.Literal["login"] | MissingType = MISSING,
        query_params: dict[str, t.Any] | None = None,
    ) -> str:
        """
        Get the authorization URL to which users should be sent.
        This method may only be called after ``oauth2_start_flow``
        has been called on this ``AuthClient``.

        :param session_required_identities: A list of identities must be
            added to the session.
        :param session_required_single_domain: A list of domain requirements
            which must be satisfied by identities added to the session.
        :param session_required_policies: A list of IDs for policies which must
            be satisfied by the user.
        :param session_required_mfa: Whether MFA is required for the session.
        :param session_message: A message to be displayed to the user by Globus Auth.
        :param prompt:
            Control whether a user is required to log in before the authorization step.

            If set to "login", the user must authenticate with an identity provider
            even if they are already logged in. Setting this parameter can help ensure
            that a user's session meets known or unknown session requirement policies
            and avoid additional login flows.
        :param query_params: Additional query parameters to include in the
            authorize URL. Primarily for internal use
        """
        if not self.current_oauth2_flow_manager:
            log.error("OutOfOrderOperations(get_authorize_url before start_flow)")
            raise exc.GlobusSDKUsageError(
                "Cannot get authorize URL until starting an OAuth2 flow. "
                "Call the oauth2_start_flow() method on this "
                "AuthClient to resolve"
            )
        query_params = {
            "session_required_identities": commajoin(session_required_identities),
            "session_required_single_domain": commajoin(session_required_single_domain),
            "session_required_policies": commajoin(session_required_policies),
            "session_required_mfa": session_required_mfa,
            "session_message": session_message,
            "prompt": prompt,
            **(query_params or {}),
        }
        auth_url = self.current_oauth2_flow_manager.get_authorize_url(
            query_params=query_params
        )
        log.debug(f"Got authorization URL: {auth_url}")
        return auth_url

    def oauth2_exchange_code_for_tokens(
        self, auth_code: str
    ) -> OAuthAuthorizationCodeResponse:
        """
        Exchange an authorization code for a token or tokens.

        :param auth_code: An auth code typically obtained by sending the user to the
            authorize URL. The code is a very short-lived credential which this method
            is exchanging for tokens. Tokens are the credentials used to authenticate
            against Globus APIs.
        """
        log.debug(
            "Final Step of 3-legged OAuth2 Flows: "
            "Exchanging authorization code for token(s)"
        )
        if not self.current_oauth2_flow_manager:
            log.error("OutOfOrderOperations(exchange_code before start_flow)")
            raise exc.GlobusSDKUsageError(
                "Cannot exchange auth code until starting an OAuth2 flow. "
                "Call the oauth2_start_flow() method on this "
                "AuthClient to resolve"
            )

        return self.current_oauth2_flow_manager.exchange_code_for_tokens(auth_code)

    def oauth2_refresh_token(
        self,
        refresh_token: str,
        *,
        body_params: dict[str, t.Any] | None = None,
    ) -> OAuthRefreshTokenResponse:
        r"""
        Exchange a refresh token for a
        :class:`OAuthTokenResponse <.OAuthTokenResponse>`, containing
        an access token.

        Does a token call of the form

        .. code-block:: none

            refresh_token=<refresh_token>
            grant_type=refresh_token

        plus any additional parameters you may specify.

        :param refresh_token: A Globus Refresh Token as a string
        :param body_params: A dict of extra params to encode in the refresh call.
        """
        log.debug("Executing token refresh; typically requires client credentials")
        form_data = {"refresh_token": refresh_token, "grant_type": "refresh_token"}
        return self.oauth2_token(
            form_data, body_params=body_params, response_class=OAuthRefreshTokenResponse
        )

    def oauth2_revoke_token(
        self,
        token: str,
        *,
        body_params: dict[str, t.Any] | None = None,
    ) -> GlobusHTTPResponse:
        """
        Revoke a token. It can be an Access Token or a Refresh token.

        This call should be used to revoke tokens issued to your client,
        rendering them inert and not further usable. Typically, this is
        incorporated into "logout" functionality, but it should also be used if
        the client detects that its tokens are in an unsafe location (e.x.
        found in a world-readable logfile).

        You can check the "active" status of the token after revocation if you
        want to confirm that it was revoked.

        :param token: The token which should be revoked
        :param body_params: Additional parameters to include in the revocation
            body, which can help speed the revocation process. Primarily for
            internal use

        **Examples**

        >>> from globus_sdk import ConfidentialAppAuthClient
        >>> ac = ConfidentialAppAuthClient(CLIENT_ID, CLIENT_SECRET)
        >>> ac.oauth2_revoke_token('<token_string>')
        """
        log.debug("Revoking token")
        body = {"token": token}

        # if this client has no way of authenticating itself but
        # it does have a client_id, we'll send that in the request
        no_authentication = self.authorizer is None or isinstance(
            self.authorizer, NullAuthorizer
        )
        if no_authentication and self.client_id:
            log.debug("Revoking token with unauthenticated client")
            body.update({"client_id": self.client_id})
        body.update(body_params or {})
        return self.post("/v2/oauth2/token/revoke", data=body, encoding="form")

    @t.overload
    def oauth2_token(
        self,
        form_data: dict[str, t.Any],
    ) -> OAuthTokenResponse: ...

    @t.overload
    def oauth2_token(
        self,
        form_data: dict[str, t.Any],
        *,
        body_params: dict[str, t.Any] | None,
    ) -> OAuthTokenResponse: ...

    @t.overload
    def oauth2_token(
        self,
        form_data: dict[str, t.Any],
        *,
        response_class: type[RT],
    ) -> RT: ...

    @t.overload
    def oauth2_token(
        self,
        form_data: dict[str, t.Any],
        *,
        body_params: dict[str, t.Any] | None,
        response_class: type[RT],
    ) -> RT: ...

    def oauth2_token(
        self,
        form_data: dict[str, t.Any],
        *,
        body_params: dict[str, t.Any] | None = None,
        response_class: type[OAuthTokenResponse] | type[RT] = OAuthTokenResponse,
    ) -> OAuthTokenResponse | RT:
        """
        This is the generic form of calling the OAuth2 Token endpoint.
        It takes ``form_data``, a dict which will be encoded in a form POST
        body on the request.

        Generally, users of the SDK should not call this method unless they are
        implementing OAuth2 flows.

        :param response_class: This is used by calls to the oauth2_token endpoint which
            need to specialize their responses. For example,
            :meth:`oauth2_get_dependent_tokens \
            <globus_sdk.ConfidentialAppAuthClient.oauth2_get_dependent_tokens>`
            requires a specialize response class to handle the dramatically different
            format of the Dependent Token Grant response
        :param form_data: The main body of the request
        :param body_params: Any additional parameters to be passed through
            as body parameters.
        :rtype: ``response_class`` if set, defaults to
            :py:attr:`~globus_sdk.OAuthTokenResponse`
        """
        log.debug("Fetching new token from Globus Auth")
        # use the fact that requests implicitly encodes the `data` parameter as
        # a form POST
        data = {**form_data, **(body_params or {})}
        return response_class(
            self.post(
                "/v2/oauth2/token",
                data=data,
                encoding="form",
            )
        )



import uuid

import pytest

from globus_sdk import MISSING
from globus_sdk._payload import GlobusPayload
from globus_sdk.transport import FormRequestEncoder, JSONRequestEncoder, RequestEncoder


@pytest.mark.parametrize("data", ("foo", b"bar"))
def test_text_request_encoder_accepts_string_data(data):
    encoder = RequestEncoder()
    request = encoder.encode(
        "GET",
        "http://bogus/foo",
        data=data,
        params={},
        headers={},
    )
    assert request.data == data


@pytest.mark.parametrize(
    "encoder_class",
    [FormRequestEncoder, JSONRequestEncoder, RequestEncoder],
)
def test_all_request_encoders_stringify_uuids_in_params_and_headers(encoder_class):
    # ensure the different classes get data which they will find palatable
    data = "bar" if encoder_class is RequestEncoder else {"bar": "baz"}

    id_ = uuid.uuid1()
    id_str = str(id_)

    encoder = encoder_class()
    request = encoder.encode(
        "GET",
        "http://bogus/foo",
        params={"id": id_},
        data=data,
        headers={"X-UUID": id_},
    )

    assert request.params["id"] == id_str
    assert request.headers["X-UUID"] == id_str


@pytest.mark.parametrize(
    "encoder_class",
    [FormRequestEncoder, JSONRequestEncoder, RequestEncoder],
)
def test_all_request_encoders_remove_missing_in_params_and_headers(encoder_class):
    # ensure the different classes get data which they will find palatable
    data = "bar" if encoder_class is RequestEncoder else {"bar": "baz"}

    encoder = encoder_class()
    request = encoder.encode(
        "GET",
        "http://bogus/foo",
        params={"foo": "foo", "bar": MISSING},
        data=data,
        headers={
            "X-FOO": "foo",
            "X-BAR": MISSING,
        },
    )
    assert request.params == {"foo": "foo"}
    if encoder_class is JSONRequestEncoder:
        assert request.headers == {"X-FOO": "foo", "Content-Type": "application/json"}
    else:
        assert request.headers == {"X-FOO": "foo"}


@pytest.mark.parametrize(
    "using_payload_type, payload_contents, expected_data",
    [
        # basic dicts
        (False, {"foo": 1}, {"foo": 1}),
        (True, {"foo": 1}, {"foo": 1}),
        # containing a UUID (stringifies)
        (True, {"foo": uuid.UUID(int=1)}, {"foo": str(uuid.UUID(int=1))}),
        (False, {"foo": uuid.UUID(int=1)}, {"foo": str(uuid.UUID(int=1))}),
        # containing MISSING (removes)
        (True, {"foo": MISSING}, {}),
        (False, {"foo": MISSING}, {}),
        # nested payload wrappers (get dictified / "unwrapped")
        (
            True,
            {"bar": GlobusPayload(foo=1), "baz": [2, GlobusPayload(foo=1)]},
            {"bar": {"foo": 1}, "baz": [2, {"foo": 1}]},
        ),
        # document with UUIDs and tuples buried inside nested structures
        (
            True,
            {"foo": ({"bar": uuid.UUID(int=2)},)},
            {"foo": [{"bar": str(uuid.UUID(int=2))}]},
        ),
    ],
)
def test_json_encoder_payload_preparation(
    using_payload_type, payload_contents, expected_data
):
    encoder = JSONRequestEncoder()
    x = GlobusPayload() if using_payload_type else {}
    for k, v in payload_contents.items():
        x[k] = v
    request = encoder.encode(
        "POST",
        "http://bogus/foo",
        params={},
        data=x,
        headers={},
    )
    assert request.json == expected_data


def test_json_encoder_is_well_defined_on_array_containing_missing():
    # this is a test for a behavior which is publicly undefined
    # (we do not advertise or encourage its use)
    #
    # it ensures that we will not crash and will behave "reasonably" in this situation
    encoder = JSONRequestEncoder()
    x = [None, 1, MISSING, 0]
    request = encoder.encode(
        "POST",
        "http://bogus/foo",
        params={},
        data=x,
        headers={},
    )
    assert request.json == [None, 1, 0]


@pytest.mark.parametrize(
    "using_payload_type, payload_contents, expected_data",
    [
        # basic dicts
        (True, {"foo": 1}, {"foo": 1}),
        (False, {"foo": 1}, {"foo": 1}),
        # containing a UUID (stringifies)
        (True, {"foo": uuid.UUID(int=1)}, {"foo": str(uuid.UUID(int=1))}),
        (False, {"foo": uuid.UUID(int=1)}, {"foo": str(uuid.UUID(int=1))}),
        # containing MISSING (removes)
        (True, {"foo": MISSING}, {}),
        (False, {"foo": MISSING}, {}),
    ],
)
def test_form_encoder_payload_preparation(
    using_payload_type, payload_contents, expected_data
):
    encoder = FormRequestEncoder()
    x = GlobusPayload() if using_payload_type else {}
    for k, v in payload_contents.items():
        x[k] = v
    request = encoder.encode(
        "POST",
        "http://bogus/foo",
        params={},
        data=x,
        headers={},
    )
    assert request.data == expected_data



import json
import uuid

from globus_sdk.experimental import TunnelCreateDocument
from globus_sdk.testing import get_last_request, load_response


def test_create_tunnel(client):
    meta = load_response(client.create_tunnel).metadata

    submission_id = uuid.uuid4()
    data = TunnelCreateDocument(
        meta["initiator_ap"],
        meta["listener_ap"],
        submission_id=submission_id,
        label=meta["display_name"],
    )

    res = client.create_tunnel(data)
    assert res.http_status == 200
    assert res["data"]["type"] == "Tunnel"

    req = get_last_request()
    sent = json.loads(req.body)
    assert (
        sent["data"]["relationships"]["initiator"]["data"]["id"] == meta["initiator_ap"]
    )
    assert (
        sent["data"]["relationships"]["listener"]["data"]["id"] == meta["listener_ap"]
    )
    assert sent["data"]["attributes"]["submission_id"] == str(submission_id)
    assert sent["data"]["attributes"]["label"] == meta["display_name"]


def test_create_tunnel_no_submission(client):
    meta = load_response(client.create_tunnel).metadata

    data = TunnelCreateDocument(
        meta["initiator_ap"], meta["listener_ap"], label=meta["display_name"]
    )

    # confirm TunnelCreateDocument auto-generated a submission_id
    generated_uuid = data["data"]["attributes"]["submission_id"]
    assert isinstance(generated_uuid, uuid.UUID)

    res = client.create_tunnel(data)
    assert res.http_status == 200

    req = get_last_request()
    sent = json.loads(req.body)
    assert sent["data"]["attributes"]["submission_id"] == str(generated_uuid)



from xml.etree import ElementTree

import pytest


def test_automethodlist_requires_an_argument(
    sphinxext, docutils_runner, register_temporary_directive
):

    register_temporary_directive("automethodlist", sphinxext.directives.AutoMethodList)

    with pytest.raises(Exception, match=r"1 argument\(s\) required, 0 supplied\."):
        docutils_runner.to_etree(".. automethodlist::")


# choose an arbitrary object from the SDK and confirm that `automethodlist`
# will render one of its public methods
# for this case, we're using `GlobusApp.login_required()`
def test_automethodlist_of_globus_app_shows_login(sphinx_runner):
    etree = sphinx_runner.to_etree(".. automethodlist:: globus_sdk.GlobusApp")

    assert etree.tag == "document"

    paragraph_element = etree.find("paragraph")
    assert paragraph_element is not None

    emphasized_text = paragraph_element.find("strong")
    assert emphasized_text is not None
    assert emphasized_text.text == "Methods"

    method_list = etree.find("bullet_list")
    assert method_list is not None
    assert b"login_required()" in ElementTree.tostring(method_list)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

RESPONSES = ResponseSet(
    groups=RegisteredResponse(
        service="auth",
        path="/v2/oauth2/token",
        method="POST",
        json=[
            {
                "scope": "urn:globus:auth:scope:groups.api.globus.org:view_my_groups_and_memberships",  # noqa: E501
                "access_token": "groupsToken",
                "token_type": "bearer",
                "expires_in": 120,
                "resource_server": "groups.api.globus.org",
            }
        ],
        metadata={
            "rs_data": {
                "groups.api.globus.org": {
                    "access_token": "groupsToken",
                    "scope": "urn:globus:auth:scope:groups.api.globus.org:view_my_groups_and_memberships",  # noqa: E501
                }
            }
        },
    ),
    groups_with_refresh_token=RegisteredResponse(
        service="auth",
        path="/v2/oauth2/token",
        method="POST",
        json=[
            {
                "scope": "urn:globus:auth:scope:groups.api.globus.org:view_my_groups_and_memberships",  # noqa: E501
                "access_token": "groupsToken",
                "refresh_token": "groupsRefreshToken",
                "token_type": "bearer",
                "expires_in": 120,
                "resource_server": "groups.api.globus.org",
            }
        ],
        metadata={
            "rs_data": {
                "groups.api.globus.org": {
                    "access_token": "groupsToken",
                    "refresh_token": "groupsRefreshToken",
                    "scope": "urn:globus:auth:scope:groups.api.globus.org:view_my_groups_and_memberships",  # noqa: E501
                }
            }
        },
    ),
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

from ._common import GROUP_ID

RESPONSES = ResponseSet(
    metadata={"group_id": GROUP_ID},
    default=RegisteredResponse(
        service="groups",
        path=f"/v2/groups/{GROUP_ID}/policies",
        method="PUT",
        json={
            "is_high_assurance": False,
            "authentication_assurance_timeout": 28800,
            "group_visibility": "private",
            "group_members_visibility": "managers",
            "join_requests": False,
            "signup_fields": ["address1"],
        },
    ),
)



from globus_sdk.testing import load_response


def test_delete_group(groups_client):
    meta = load_response(groups_client.delete_group).metadata

    res = groups_client.delete_group(group_id=meta["group_id"])
    assert res.http_status == 200
    assert "Claptrap" in res["name"]



from .collection import DynamicScopeCollection, ScopeCollection, StaticScopeCollection
from .data import (
    AuthScopes,
    ComputeScopes,
    FlowsScopes,
    GCSCollectionScopes,
    GCSEndpointScopes,
    GroupsScopes,
    NexusScopes,
    SearchScopes,
    SpecificFlowScopes,
    TimersScopes,
    TransferScopes,
)
from .errors import ScopeCycleError, ScopeParseError
from .parser import ScopeParser
from .representation import Scope

__all__ = (
    "ScopeCollection",
    "StaticScopeCollection",
    "DynamicScopeCollection",
    "Scope",
    "ScopeParser",
    "ScopeParseError",
    "ScopeCycleError",
    "GCSCollectionScopes",
    "GCSEndpointScopes",
    "AuthScopes",
    "ComputeScopes",
    "FlowsScopes",
    "SpecificFlowScopes",
    "GroupsScopes",
    "NexusScopes",
    "SearchScopes",
    "TimersScopes",
    "TransferScopes",
)



import typing as t
import uuid

from responses.matchers import json_params_matcher

from globus_sdk.testing.models import RegisteredResponse, ResponseSet

_COMMON_RESPONSE_RECORD: dict[str, t.Any] = {
    "fqdns": [],
    "links": {"privacy_policy": None, "terms_and_conditions": None},
    "parent_client": None,
    "preselect_idp": None,
    "prompt_for_named_grant": True,
    "redirect_uris": [],
    "required_idp": None,
    "scopes": [],
    "userinfo_from_effective_identity": True,
}

PUBLIC_CLIENT_RESPONSE_RECORD: dict[str, t.Any] = {
    "grant_types": ["authorization_code", "refresh_token"],
    **_COMMON_RESPONSE_RECORD,
}

PRIVATE_CLIENT_RESPONSE_RECORD: dict[str, t.Any] = {
    "grant_types": [
        "authorization_code",
        "client_credentials",
        "refresh_token",
        "urn:globus:auth:grant_type:dependent_token",
    ],
    **_COMMON_RESPONSE_RECORD,
}

PUBLIC_CLIENT_REQUEST_ARGS: dict[str, t.Any] = {
    "name": "FOO",
    "visibility": "public",
}


def register_response(
    args: t.Mapping[str, t.Any],
) -> RegisteredResponse:
    request_args = {
        **PUBLIC_CLIENT_REQUEST_ARGS,
        **args,
    }
    # Default to public_client=True
    if "public_client" not in args and "client_type" not in args:
        request_args["public_client"] = True

    # The request body follows almost directly from the request args. Some name of args
    # to create_client() have differenlty named fields.
    request_body: t.Dict[str, t.Any] = {}
    for arg_name in request_args:
        if arg_name == "terms_and_conditions" or arg_name == "privacy_policy":
            request_body["links"] = {
                arg_name: request_args[arg_name],
                **request_body.get("links", {}),
            }
        else:
            request_body[arg_name] = request_args[arg_name]

    client_response_record = {}
    if (
        request_args.get("public_client", False)
        or request_args.get("client_type") == "public_installed_client"
    ):
        client_response_record = {
            **PUBLIC_CLIENT_RESPONSE_RECORD,
            **request_body,
        }
        if "client_type" not in client_response_record:
            client_response_record["client_type"] = "public_installed_client"
    else:
        client_response_record = {
            **PRIVATE_CLIENT_RESPONSE_RECORD,
            **request_body,
        }
        if "client_type" not in client_response_record:
            client_response_record["client_type"] = (
                "hybrid_confidential_client_resource_server"
            )

    return RegisteredResponse(
        service="auth",
        method="POST",
        path="/v2/api/clients",
        json={"client": client_response_record},
        metadata={
            # Test functions use 'args' to form request
            "args": request_args,
            # Test functions use 'response' to verify response
            "response": client_response_record,
        },
        match=[json_params_matcher({"client": request_body})],
    )


RESPONSES = ResponseSet(
    default=register_response({}),
    name=register_response({"name": str(uuid.uuid4()).replace("-", "")}),
    public_client=register_response({"public_client": True}),
    private_client=register_response({"public_client": False}),
    publicly_visible=register_response({"visibility": "public"}),
    not_publicly_visible=register_response({"visibility": "private"}),
    redirect_uris=register_response({"redirect_uris": ["https://foo.com"]}),
    links=register_response(
        {
            "terms_and_conditions": "https://foo.org",
            "privacy_policy": "https://boo.org",
        }
    ),
    required_idp=register_response({"required_idp": str(uuid.uuid1())}),
    preselect_idp=register_response({"preselect_idp": str(uuid.uuid1())}),
    client_type_confidential_client=register_response(
        {"client_type": "confidential_client"}
    ),
    client_type_public_installed_client=register_response(
        {"client_type": "public_installed_client"}
    ),
    client_type_client_identity=register_response({"client_type": "client_identity"}),
    client_type_resource_server=register_response({"client_type": "resource_server"}),
    client_type_globus_connect_server=register_response(
        {"client_type": "globus_connect_server"}
    ),
    client_type_hybrid_confidential_client_resource_server=register_response(
        {"client_type": "hybrid_confidential_client_resource_server"}
    ),
    client_type_public_webapp_client=register_response(
        {"client_type": "public_webapp_client"}
    ),
)



import pickle

import pytest

import globus_sdk


@pytest.fixture
def auth_client():
    client = globus_sdk.AuthClient()

    yield client


@pytest.fixture
def token_response(auth_client, make_oauth_token_response):
    return make_oauth_token_response(client=auth_client)


def test_pickle_and_unpickle_no_usage(token_response):
    """
    Test pickle and unpickle, with no usage of the result,
    for all pickle protocol versions supported by the current interpreter.
    """
    pickled_versions = [
        pickle.dumps(token_response, protocol=n)
        for n in range(pickle.HIGHEST_PROTOCOL + 1)
    ]
    unpickled_versions = [pickle.loads(x) for x in pickled_versions]
    for x in unpickled_versions:
        assert x.by_resource_server == token_response.by_resource_server



import globus_sdk
from globus_sdk.scopes import TransferScopes

CLIENT_ID = "61338d24-54d5-408f-a10d-66c06b59f6d2"
auth_client = globus_sdk.NativeAppAuthClient(CLIENT_ID)

# requested_scopes specifies a list of scopes to request
# instead of the defaults, only request access to the Transfer API
auth_client.oauth2_start_flow(requested_scopes=TransferScopes.all)
authorize_url = auth_client.oauth2_get_authorize_url()
print(f"Please go to this URL and login:\n\n{authorize_url}\n")

auth_code = input("Please enter the code here: ").strip()
tokens = auth_client.oauth2_exchange_code_for_tokens(auth_code)
transfer_tokens = tokens.by_resource_server["transfer.api.globus.org"]

# construct an AccessTokenAuthorizer and use it to construct the
# TransferClient
transfer_client = globus_sdk.TransferClient(
    authorizer=globus_sdk.AccessTokenAuthorizer(transfer_tokens["access_token"])
)

# Replace these with your own collection UUIDs
source_collection_id = "..."
dest_collection_id = "..."

# create a Transfer task consisting of one or more items
task_data = globus_sdk.TransferData(
    source_endpoint=source_collection_id, destination_endpoint=dest_collection_id
)
task_data.add_item(
    "/share/godata/file1.txt",  # source
    "/~/minimal-example-transfer-script-destination.txt",  # dest
)

# submit, getting back the task ID
task_doc = transfer_client.submit_transfer(task_data)
task_id = task_doc["task_id"]
print(f"submitted transfer, task_id={task_id}")



"""
Tests for TransferClient.operation_stat
"""

import urllib.parse

from globus_sdk.testing import get_last_request, load_response


def test_operation_stat(client):
    meta = load_response(client.operation_stat).metadata
    endpoint_id = meta["endpoint_id"]
    path = "/home/share/godata/file1.txt"
    res = client.operation_stat(endpoint_id, path)

    assert res["name"] == "file1.txt"
    assert res["type"] == "file"
    assert res["size"] == 4

    req = get_last_request()
    parsed_qs = urllib.parse.parse_qs(urllib.parse.urlparse(req.url).query)
    assert parsed_qs == {"path": [path]}



import importlib
import sys
import typing as t

from .api import ErrorSubdocument, GlobusAPIError
from .base import GlobusError, GlobusSDKUsageError, ValidationError
from .err_info import (
    AuthorizationParameterInfo,
    ConsentRequiredInfo,
    ErrorInfo,
    ErrorInfoContainer,
)
from .warnings import RemovedInV5Warning, warn_deprecated

__all__ = (
    "GlobusError",
    "GlobusSDKUsageError",
    "ValidationError",
    "GlobusAPIError",
    "ErrorSubdocument",
    "NetworkError",
    "GlobusTimeoutError",
    "GlobusConnectionTimeoutError",
    "GlobusConnectionError",
    "convert_request_exception",
    "ErrorInfo",
    "ErrorInfoContainer",
    "AuthorizationParameterInfo",
    "ConsentRequiredInfo",
    "RemovedInV5Warning",
    "warn_deprecated",
)

# imports from `globus_sdk.exc.convert` are done lazily
#
# this ensures that we do not eagerly import `requests` when attempting to use SDK
# components which do not need it, but which do need errors (e.g., RemovedInV5Warning)
# and we avoid paying the performance penalty for importing the relevant dependencies
if t.TYPE_CHECKING:
    from .convert import (
        GlobusConnectionError,
        GlobusConnectionTimeoutError,
        GlobusTimeoutError,
        NetworkError,
        convert_request_exception,
    )
else:
    _LAZY_IMPORT_TABLE = {
        "convert": {
            "GlobusConnectionError",
            "GlobusConnectionTimeoutError",
            "GlobusTimeoutError",
            "NetworkError",
            "convert_request_exception",
        }
    }

    def __getattr__(name: str) -> t.Any:
        for modname, items in _LAZY_IMPORT_TABLE.items():
            if name in items:
                mod = importlib.import_module("." + modname, __name__)
                value = getattr(mod, name)
                setattr(sys.modules[__name__], name, value)
                return value

        raise AttributeError(f"module {__name__} has no attribute {name}")



from globus_sdk.testing import load_response


def test_get_run_logs(flows_client):
    metadata = load_response(flows_client.get_run_logs).metadata

    resp = flows_client.get_run_logs(metadata["run_id"])
    assert resp.http_status == 200


def test_get_run_logs_paginated(flows_client):
    metadata = load_response(flows_client.get_run_logs, case="paginated").metadata

    paginator = flows_client.paginated.get_run_logs(metadata["run_id"])
    responses = list(paginator.pages())

    assert len(responses) == 2
    assert len(responses[0]["entries"]) == 10
    assert len(responses[1]["entries"]) == 2


def test_get_run_logs_manually_paginated(flows_client):
    metadata = load_response(flows_client.get_run_logs, case="paginated").metadata

    resp = flows_client.get_run_logs(metadata["run_id"])
    assert resp.http_status == 200
    assert resp["has_next_page"]
    assert "marker" in resp.data
    assert len(resp["entries"]) == 10

    resp2 = flows_client.get_run_logs(metadata["run_id"], marker=resp["marker"])
    assert resp2.http_status == 200
    assert not resp2["has_next_page"]
    assert len(resp2["entries"]) == 2



from responses.matchers import query_param_matcher

from globus_sdk.testing import RegisteredResponse, ResponseList, ResponseSet

RUN_ID = "cfdaf0a4-0931-40af-b974-b619ce69f401"
OWNER_URN = "urn:globus:auth:identity:944cfbe8-60f8-474d-a634-a0c1ad543a54"
RUN_LOGS_RESPONSE = {
    "limit": 10,
    "has_next_page": False,
    "entries": [
        {
            "time": "2023-04-25T18:54:30.683000+00:00",
            "code": "FlowStarted",
            "description": "The Flow Instance started execution",
            "details": {"input": {}},
        },
        {
            "time": "2023-04-25T18:54:30.715000+00:00",
            "code": "ActionStarted",
            "description": "State SyncHelloWorld of type Action started",
            "details": {
                "state_name": "SyncHelloWorld",
                "state_type": "Action",
                "input": {"echo_string": "sync!"},
            },
        },
        {
            "time": "2023-04-25T18:54:31.850000+00:00",
            "code": "ActionCompleted",
            "description": "State SyncHelloWorld of type Action completed",
            "details": {
                "state_name": "SyncHelloWorld",
                "state_type": "Action",
                "output": {
                    "action_id": "19NqhOnDlt2Y3",
                    "completion_time": "2023-04-25T18:54:31.341170+00:00",
                    "creator_id": OWNER_URN,
                    "details": {"Hello": "World", "hello": "sync!"},
                    "display_status": "SUCCEEDED",
                    "label": None,
                    "manage_by": [],
                    "monitor_by": [],
                    "release_after": None,
                    "start_time": "2023-04-25T18:54:31.340484+00:00",
                    "status": "SUCCEEDED",
                    "state_name": "SyncHelloWorld",
                },
            },
        },
        {
            "time": "2023-04-25T18:54:31.913000+00:00",
            "code": "FlowSucceeded",
            "description": "The Flow Instance completed successfully",
            "details": {
                "output": {
                    "action_id": "19NqhOnDlt2Y3",
                    "completion_time": "2023-04-25T18:54:31.341170+00:00",
                    "creator_id": OWNER_URN,
                    "details": {"Hello": "World", "hello": "sync!"},
                    "display_status": "SUCCEEDED",
                    "label": None,
                    "manage_by": [],
                    "monitor_by": [],
                    "release_after": None,
                    "start_time": "2023-04-25T18:54:31.340484+00:00",
                    "status": "SUCCEEDED",
                    "state_name": "SyncHelloWorld",
                }
            },
        },
    ],
}

PAGINATED_RUN_LOG_RESPONSES = [
    {
        "limit": 10,
        "has_next_page": True,
        "marker": "fake_run_logs_marker",
        "entries": [
            {
                "time": "2023-04-25T18:54:30.683000+00:00",
                "code": "FlowStarted",
                "description": "The Flow Instance started execution",
                "details": {"input": {}},
            },
            {
                "time": "2023-04-25T18:54:30.715000+00:00",
                "code": "PassStarted",
                "description": "State PassState of type Pass started",
                "details": {
                    "state_name": "PassState",
                    "state_type": "Pass",
                    "input": {},
                },
            },
            {
                "time": "2023-04-25T18:54:30.715000+00:00",
                "code": "PassCompleted",
                "description": "State PassState of type Pass completed",
                "details": {
                    "state_name": "PassState",
                    "state_type": "Pass",
                    "output": {},
                },
            },
            {
                "time": "2023-04-25T18:54:30.715000+00:00",
                "code": "PassStarted",
                "description": "State PassState2 of type Pass started",
                "details": {
                    "state_name": "PassState2",
                    "state_type": "Pass",
                    "input": {},
                },
            },
            {
                "time": "2023-04-25T18:54:30.715000+00:00",
                "code": "PassCompleted",
                "description": "State PassState2 of type Pass completed",
                "details": {
                    "state_name": "PassState2",
                    "state_type": "Pass",
                    "output": {},
                },
            },
            {
                "time": "2023-04-25T18:54:30.715000+00:00",
                "code": "PassStarted",
                "description": "State PassState3 of type Pass started",
                "details": {
                    "state_name": "PassState3",
                    "state_type": "Pass",
                    "input": {},
                },
            },
            {
                "time": "2023-04-25T18:54:30.715000+00:00",
                "code": "PassCompleted",
                "description": "State PassState3 of type Pass completed",
                "details": {
                    "state_name": "PassState3",
                    "state_type": "Pass",
                    "output": {},
                },
            },
            {
                "time": "2023-04-25T18:54:30.715000+00:00",
                "code": "PassStarted",
                "description": "State PassState4 of type Pass started",
                "details": {
                    "state_name": "PassState4",
                    "state_type": "Pass",
                    "input": {},
                },
            },
            {
                "time": "2023-04-25T18:54:30.715000+00:00",
                "code": "PassCompleted",
                "description": "State PassState4 of type Pass completed",
                "details": {
                    "state_name": "PassState4",
                    "state_type": "Pass",
                    "output": {},
                },
            },
            {
                "time": "2023-04-25T18:54:30.715000+00:00",
                "code": "ActionStarted",
                "description": "State SyncHelloWorld of type Action started",
                "details": {
                    "state_name": "SyncHelloWorld",
                    "state_type": "Action",
                    "input": {"echo_string": "sync!"},
                },
            },
        ],
    },
    {
        "limit": 10,
        "has_next_page": False,
        "entries": [
            {
                "time": "2023-04-25T18:54:31.850000+00:00",
                "code": "ActionCompleted",
                "description": "State SyncHelloWorld of type Action completed",
                "details": {
                    "state_name": "SyncHelloWorld",
                    "state_type": "Action",
                    "output": {
                        "action_id": "19NqhOnDlt2Y3",
                        "completion_time": "2023-04-25T18:54:31.341170+00:00",
                        "creator_id": OWNER_URN,
                        "details": {"Hello": "World", "hello": "sync!"},
                        "display_status": "SUCCEEDED",
                        "label": None,
                        "manage_by": [],
                        "monitor_by": [],
                        "release_after": None,
                        "start_time": "2023-04-25T18:54:31.340484+00:00",
                        "status": "SUCCEEDED",
                        "state_name": "SyncHelloWorld",
                    },
                },
            },
            {
                "time": "2023-04-25T18:54:31.913000+00:00",
                "code": "FlowSucceeded",
                "description": "The Flow Instance completed successfully",
                "details": {
                    "output": {
                        "action_id": "19NqhOnDlt2Y3",
                        "completion_time": "2023-04-25T18:54:31.341170+00:00",
                        "creator_id": OWNER_URN,
                        "details": {"Hello": "World", "hello": "sync!"},
                        "display_status": "SUCCEEDED",
                        "label": None,
                        "manage_by": [],
                        "monitor_by": [],
                        "release_after": None,
                        "start_time": "2023-04-25T18:54:31.340484+00:00",
                        "status": "SUCCEEDED",
                        "state_name": "SyncHelloWorld",
                    }
                },
            },
        ],
    },
]


RESPONSES = ResponseSet(
    metadata={"run_id": RUN_ID},
    default=RegisteredResponse(
        service="flows",
        method="GET",
        path=f"/runs/{RUN_ID}/log",
        json=RUN_LOGS_RESPONSE,
    ),
    paginated=ResponseList(
        RegisteredResponse(
            service="flows",
            method="GET",
            path=f"/runs/{RUN_ID}/log",
            json=PAGINATED_RUN_LOG_RESPONSES[0],
            match=[query_param_matcher(params={})],
        ),
        RegisteredResponse(
            service="flows",
            method="GET",
            path=f"/runs/{RUN_ID}/log",
            json=PAGINATED_RUN_LOG_RESPONSES[1],
            match=[
                query_param_matcher(
                    params={"marker": PAGINATED_RUN_LOG_RESPONSES[0]["marker"]}
                )
            ],
        ),
    ),
)



from .personal import GlobusConnectPersonalOwnerInfo, LocalGlobusConnectPersonal
from .server import LocalGlobusConnectServer

__all__ = (
    "GlobusConnectPersonalOwnerInfo",
    "LocalGlobusConnectPersonal",
    "LocalGlobusConnectServer",
)



from globus_sdk.testing.models import RegisteredResponse, ResponseSet

DEFAULT_RESPONSE_DOC = {
    "queue_prefix": "some_prefix",
    "connection_url": "amqps://user:password@amqp.fqdn",
}

RESPONSES = ResponseSet(
    default=RegisteredResponse(
        service="compute",
        path="/v2/get_amqp_result_connection_url",
        method="GET",
        json=DEFAULT_RESPONSE_DOC,
    ),
)



from __future__ import annotations

import typing as t
from pydoc import locate

from docutils.parsers.rst import directives

from globus_sdk.scopes import Scope, ScopeCollection

from .add_content_directive import AddContentDirective


class ListKnownScopes(AddContentDirective):
    has_content = False
    required_arguments = 1
    optional_arguments = 0
    option_spec = {
        "example_scope": directives.unchanged,
        # Allow overriding the base name to match how the ScopeCollection will
        # be accessed.
        "base_name": directives.unchanged,
    }

    def gen_rst(self) -> t.Iterator[str]:
        sc_name = self.arguments[0]
        sc_basename = sc_name.split(".")[-1]
        if "base_name" in self.options:
            sc_basename = self.options["base_name"]

        example_scope = None
        if "example_scope" in self.options:
            example_scope = self.options["example_scope"].strip()
        known_scopes = extract_known_scopes(sc_name)
        if example_scope is None:
            example_scope = known_scopes[0]

        yield ""
        yield "Various scopes are available as attributes of this object."
        yield f"For example, access the ``{example_scope}`` scope with"
        yield ""
        yield f">>> {sc_basename}.{example_scope}"
        yield ""
        yield "**Supported Scopes**"
        yield ""
        for s in known_scopes:
            yield f"* ``{s}``"
        yield ""
        yield ""


def extract_known_scopes(scope_collection_name: str) -> list[str]:
    sc = locate(scope_collection_name)
    if isinstance(sc, ScopeCollection):
        return [name for name in dir(sc) if isinstance(getattr(sc, name), Scope)]

    raise RuntimeError(
        f"Expected {sc} to be a scope collection, but got {type(sc)} instead"
    )
